import os
import wx
import wx.lib.agw.flatnotebook as fnb
import abipy.gui.awx as awx

from collections import OrderedDict
from monty.string import list_strings, is_string
from abipy.flowtk import EventsParser


class AbinitEventsPanel(awx.Panel):
    """
    Panel with a TreeCtrl that allows the user to navigate
    the events (WARNINGS/COMMENTS/ERRORS) reported by ABINIT
    in the main output file or in the log file.
    """
    def __init__(self, parent, filepath, **kwargs):
        super(AbinitEventsPanel, self).__init__(parent, -1, **kwargs)

        self.filepath = os.path.abspath(filepath)

        parser = EventsParser()
        self.events = events = parser.parse(self.filepath)

        main_sizer = wx.BoxSizer(wx.VERTICAL)
        vbox = wx.BoxSizer(wx.VERTICAL)
        panel1 = awx.Panel(self, -1)
        panel2 = awx.Panel(self, -1)

        self.tree = tree = wx.TreeCtrl(panel1, 1, wx.DefaultPosition, (-1, -1), wx.TR_HIDE_ROOT | wx.TR_HAS_BUTTONS)

        root = self.tree.AddRoot('Events')

        err_tree = tree.AppendItem(root, "%d Errors" % events.num_errors)
        warn_tree = tree.AppendItem(root, "%d Warnings" % events.num_warnings)
        com_tree = tree.AppendItem(root, "%d Comments" % events.num_comments)

        for e in self.events.errors:
            tree.AppendItem(err_tree, "line: " + str(e.lineno), data=wx.TreeItemData(e.message))

        for w in self.events.warnings:
            tree.AppendItem(warn_tree, "line: " + str(w.lineno), data=wx.TreeItemData(w.message))

        for c in self.events.comments:
            tree.AppendItem(com_tree, "line: " + str(c.lineno), data=wx.TreeItemData(c.message))

        tree.Bind(wx.EVT_TREE_SEL_CHANGED, self.OnSelChanged, id=1)

        self.display = wx.StaticText(panel2, -1, '', (10, 10), style=wx.ALIGN_LEFT)

        vbox.Add(self.tree, 1, wx.EXPAND)
        main_sizer.Add(panel1, 1, wx.EXPAND)
        main_sizer.Add(panel2, 1, wx.EXPAND)
        panel1.SetSizerAndFit(vbox)

        self.SetSizerAndFit(main_sizer)
        self.Centre()

    @property
    def has_events(self):
        return len(self.events) > 0

    def OnSelChanged(self, event):
        item = event.GetItem()
        proxy = self.tree.GetItemData(item)
        if proxy is None: return
        data = proxy.GetData()
        self.display.SetLabel(data)


class AbinitEventsFrame(awx.Frame):
    """
    Frame with an EventsPanel
    """
    def __init__(self, parent, filepath, **kwargs):
        filepath = os.path.abspath(filepath)

        if "title" not in kwargs:
            kwargs["title"] = "Abinit Events: %s" % os.path.basename(filepath)

        super(AbinitEventsFrame, self).__init__(parent, **kwargs)

        self.event_panel = AbinitEventsPanel(self, filepath)


class AbiOutLogDirCtrl(wx.GenericDirCtrl):

    def __init__(self, *args, **kwargs):
        if "filter" not in kwargs:
            kwargs["filter"] = "All files (*.*)|*.*|ABINIT output files (*.abo)|*.out| ABINIT log files (*.log)|*.log"
        if "dir" not in kwargs:
            kwargs["dir"] = os.getcwd()
        if "style" not in kwargs:
            kwargs["style"] = wx.TR_MULTIPLE

        super(AbiOutLogDirCtrl, self).__init__(*args, **kwargs)

        self.Bind(wx.EVT_TREE_ITEM_ACTIVATED, self.OnItemActivated)
        #self.Bind(wx.EVT_TREE_ITEM_RIGHT_CLICK, self.OnRightClick)

    def OnItemActivated(self, event):
        path = self.GetFilePath()
        if not path: return
        EventFrame(self, path).Show()

    #def OnRightClick(self, event):
    #    path = self.GetFilePath()
    #    if not path: return
    #    print("in right with path %s" % path)


class AbinitEventsNotebookFrame(awx.Frame):
    """
    This frame receives a list of filenames and displays the ABINIT events in a notebook.
    """
    def __init__(self, parent, filenames, num_dirs=2, **kwargs):
        """
        Args:
            parent:
                Parent Widget.
            filenames:
                List of filenames.
            num_dirs:
                Maximum number of directories that will be shown in the tab.
        """
        if "title" not in kwargs:
            kwargs["title"] = "Abinit Events"

        super(AbinitEventsNotebookFrame, self).__init__(parent, **kwargs)

        filenames = list_strings(filenames)

        # Remove inexistent files.
        filenames = filter(os.path.exists, filenames)

        if not filenames:
            return

        # Here we create a panel and a notebook on the panel
        panel = awx.Panel(self)

        nb = fnb.FlatNotebook(panel)

        for fname in filenames:
            page = AbinitEventsPanel(nb, fname)
            page_name = fname

            if num_dirs > 0:
                tokens = page_name.split(os.path.sep)
                page_name = os.path.join(*tokens[-num_dirs:])

            # Add only files for which we have events.
            #if page.has_events:

            # Add the pages to the notebook with the name to show on the tab
            nb.AddPage(page, text=page_name)

        # Finally, put the notebook in a sizer for the panel to manage the layout
        sizer = wx.BoxSizer()
        sizer.Add(nb, 1, wx.EXPAND)

        panel.SetSizerAndFit(sizer)


def wxapp_events(root):
    """
    Start up the AbinitOutputViewer application.

    Args:
        root:
            Can be: None, filename, directory name or list of filenames.
            None means that we just open the browser without accessing any file.
            If root is a directory, we locate all the output files
            starting from root and we visualize them in the main Frame.
    """
    if root is None:
        filenames = []

    elif is_string(root):
        root = os.path.abspath(root)
        if os.path.isdir(root):
            filenames = [os.path.join(root, f) for f in os.listdir(root) if f.endswith(".abo")]
        else:
            filenames = [root]
    else:
        filenames = root

    class AbiEventsViewerApp(awx.App):
        def OnInit(self):
            frame = AbinitEventsNotebookFrame(None, filenames)
            self.SetTopWindow(frame)
            frame.Show()
            return True

    return AbiEventsViewerApp()



import wx

from collections import OrderedDict
from abipy.gui import awx

from abipy.tools.fftprof import FFTProf
from abipy.gui.editor import SimpleTextViewer
from .awx.panels import LinspaceControl


FFTALGS = OrderedDict([
    (112, "Goedecker"),
    (312, "FFTW3"),
    (412, "Goedecker-2002"),
    (512, "MKL-DFTI"),
])


class FftalgsPanel(wx.Panel):
    def __init__(self, parent):
        wx.Panel.__init__(self, parent, id=-1)

        main_sizer = wx.BoxSizer(wx.VERTICAL)
        static_sizer = wx.StaticBoxSizer(wx.StaticBox(self, -1, "FFT Libraries"), wx.VERTICAL)
        grid_sizer = wx.GridSizer(0, 2, 0, 0)

        self.check_boxes = OrderedDict()

        for fftalg, fft_name in FFTALGS.items():
            cbox = wx.CheckBox(self, -1, fft_name)
            if fftalg == 112:
                cbox.SetValue(True)
            else:
                cbox.SetValue(False)

            self.check_boxes[fftalg] = cbox
            grid_sizer.Add(cbox, 0, wx.ALL, 5)

        static_sizer.Add(grid_sizer, 0, wx.ALL | wx.EXPAND, 5)
        main_sizer.Add(static_sizer, 1, wx.EXPAND, 5)

        self.SetSizerAndFit(main_sizer)

    def GetValues(self):
        """Returns a list with the fftalgs selected by the user."""
        fftalgs = []
        for (fftalg, cbox) in self.check_boxes.items():
            if cbox.GetValue():
                fftalgs.append(fftalg)

        if not fftalgs:
            fftalgs = [112]

        return fftalgs


class ControlPanel(wx.Panel):
    """
    !!   &CONTROL
    !!     tasks = string specifying the tasks to perform i.e. the routines that should be tested or profiled.
    !!             allowed values:
    !!                 fourdp --> Test FFT transforms of density and potentials on the full box.
    !!                 fourwf --> Test FFT transforms of wavefunctions using the zero-padded algorithm.
    !!                 gw_fft --> Test the FFT transforms used in the GW code.
    !!                 all    --> Test all FFT routines (DEFAULT)
    !!     fftalgs = list of fftalg values (used to select the FFT libraries to use, see abinit doc for more info)
    !!     ncalls = integer defining the number of calls for each tests. The final Wall time and CPU time
    !!              are computed by averaging the final results over ncalls executions.
    !!     max_nthreads = Maximum number of threads (DEFAULT 1, meaningful only if the code
    !!                    uses threaded external libraries or OpenMP parallelization)
    !!     ndat   = integer specifying how many FFT transforms should be executed for each call to the FFT routine
    !!              (same meaning as the ndat input variable passed to fourwf)
    !!     necut  = Used if tasks = "bench". Specifies the number of cutoff energies to profile (see also ecut_arth)
    !!     ecut_arth = Used if tasks = "bench". Used to generate an arithmetic progression of cutoff energies
    !!                 whose starting value is ecut_arth(1) and whose step is ecut_arth(2)
    """
    BENCHMARKS = ["bench_fourwf", "bench_fourdp", "bench_rhotwg"]

    def __init__(self, parent):
        wx.Panel.__init__(self, parent, id=-1)

        main_sizer = wx.BoxSizer(wx.VERTICAL)

        hsz1 = wx.BoxSizer(wx.HORIZONTAL)

        text = wx.StaticText(self, -1, "Tasks:")
        text.Wrap(-1)
        text.SetToolTipString("Select the routines for the benchmark.")
        self.task_choice = wx.Choice(self, -1, wx.DefaultPosition, wx.DefaultSize, self.BENCHMARKS, 0)
        self.task_choice.SetSelection(0)

        hsz1.Add(text, 0, wx.ALIGN_CENTER_VERTICAL | wx.TOP | wx.BOTTOM | wx.LEFT, 5)
        hsz1.Add(self.task_choice, 0, wx.ALIGN_CENTER_VERTICAL | wx.ALL, 5)

        text = wx.StaticText(self, -1, "ndat:")
        text.Wrap(-1)
        text.SetToolTipString("Number of FFT transform per call.")
        self.ndat_ctrl = wx.SpinCtrl(self, -1, value="1", min=1)

        hsz1.Add(text, 0, wx.ALIGN_CENTER_VERTICAL | wx.TOP | wx.BOTTOM | wx.LEFT, 5)
        hsz1.Add(self.ndat_ctrl, 0, wx.ALIGN_CENTER_VERTICAL | wx.ALL, 5)

        text = wx.StaticText(self, -1, "ncalls:")
        text.Wrap(-1)
        text.SetToolTipString("Number of calls (used to get mean results)")
        self.ncalls_ctrl = wx.SpinCtrl(self, -1, value="5", min=1)

        hsz1.Add(text, 0, wx.ALIGN_CENTER_VERTICAL | wx.TOP | wx.BOTTOM | wx.LEFT, 5)
        hsz1.Add(self.ncalls_ctrl, 0, wx.ALIGN_CENTER_VERTICAL | wx.ALL, 5)

        text = wx.StaticText(self, -1, "max_nthreads:")
        text.Wrap(-1)
        text.SetToolTipString("Maximum number of OpenMP threads")
        self.max_nthreads_ctrl = wx.SpinCtrl(self, -1, value="1", min=1)

        hsz1.Add(text, 0, wx.ALIGN_CENTER_VERTICAL | wx.TOP | wx.BOTTOM | wx.LEFT, 5)
        hsz1.Add(self.max_nthreads_ctrl, 0, wx.ALIGN_CENTER_VERTICAL | wx.ALL, 5)

        main_sizer.Add(hsz1, 1, wx.EXPAND, 5)

        # Control to read the range of cutoff energies.
        self.ecut_linspace = LinspaceControl(self, start=dict(value="20"), stop=dict(value="120"), num=dict(value="10"))

        static_sizer = wx.StaticBoxSizer(wx.StaticBox(self, -1, "Ecut list:"), wx.VERTICAL)
        static_sizer.Add(self.ecut_linspace, 0, wx.ALL | wx.EXPAND, 5)

        # Control to specify fftalgs.
        self.fftalgs = FftalgsPanel(self)

        #vsz1 = wx.BoxSizer(wx.VERTICAL)
        #vsz1.Add(static_sizer, 0, wx.ALIGN_CENTER_VERTICAL | wx.ALL, 5)
        #vsz1.Add(self.fftalgs, 0, wx.ALIGN_CENTER_VERTICAL | wx.ALL, 5)
        #main_sizer.Add(vsz1, 1, wx.EXPAND, 5)

        main_sizer.Add(static_sizer, 1, wx.EXPAND, 5)
        main_sizer.Add(self.fftalgs, 1, wx.EXPAND, 5)

        self.SetSizerAndFit(main_sizer)

    def GetNamelist(self):
        """
        Build and returns a string with whe &CONTROL Fortran NAMELIST.
        """
        ecut_list = self.ecut_linspace.getValues()
        if len(ecut_list) <= 1:
            return awx.showErrorMessage(self, message="Ecut list is empty or contains only one item!")
        step = ecut_list[1] - ecut_list[0]

        d = OrderedDict(
            tasks="'" + self.task_choice.GetStringSelection() + "'",
            fftalgs=",".join(str(v) for v in self.fftalgs.GetValues()),
            ncalls=self.ncalls_ctrl.GetValue(),
            max_nthreads=self.max_nthreads_ctrl.GetValue(),
            ndat=self.ndat_ctrl.GetValue(),
            necut=len(ecut_list),
            ecut_arth="%s, %s" % (ecut_list[0], step)
        )

        namelist = ["&CONTROL"]
        for k, v in d.items():
            namelist.append(k + " = " + str(v) + ",")
        namelist.append("/")

        return "\n".join(namelist)


class KpointChoice(wx.Choice):
    def __init__(self, parent):
        choices = map(str, [
            (0.1, 0.2, 0.3),
            (0.0, 0.0, 0.0),
            (1 / 2, 0, 0),
            (0, 0, 1 / 2),
            (1 / 2, 0, 1 / 2),
            (0, 1 / 2, 0),
            (1 / 2, 1 / 2, 0),
            (0, 1 / 2, 1 / 2),
            (1 / 2, 1 / 2, 1 / 2),
        ])

        wx.Choice.__init__(self, parent, -1, choices=choices)
        self.SetSelection(0)

    def GetStringSelection(self):
        s = super(KpointChoice, self).GetStringSelection()
        return s.replace("(", "").replace(")", "")


class SystemPanel(wx.Panel):
    """
    !!     ecut = cutoff energy for wavefunctions (real, Hartree units)
    !!     rprimd = Direct lattice vectors in Bohr. (3,3) matrix in Fortran column-major order
    !!     kpoint = real(3) vector specifying the reduced coordinates of the k-point of the wavefunction (used if tasks = "fourwf").
    !!               The value of the k-point defines the storage scheme (istwfk) of the u(G) coefficients and therefore
    !!               the FFT algorithm used to perform the transform u(G) <--> u(R) in fourwf.
    !!     osc_ecut = cutoff energy (Hartree) for the oscillator matrix elements computed in the GW code
    !!                 Corresponds to the input variables (ecuteps, ecutsigx) used in the main code.
    !!     nsym     =Number of symmetry operations (DEFAULT 1)
    !!     symrel(3,3,nsym) = Symmetry operation in real space used to select the FFT mesh in the routine getng (default: Identity matrix)
    """

    def __init__(self, parent):
        wx.Panel.__init__(self, parent, id=-1)

        main_sizer = wx.BoxSizer(wx.VERTICAL)

        hsz1 = wx.BoxSizer(wx.HORIZONTAL)

        text = wx.StaticText(self, -1, "kpoint:")
        text.Wrap(-1)
        text.SetToolTipString("K-point used in the FFT of the wavefunctions (istwfk).")
        self.kpoint_choice = KpointChoice(self)

        hsz1.Add(text, 0, wx.ALIGN_CENTER_VERTICAL | wx.TOP | wx.BOTTOM | wx.LEFT, 5)
        hsz1.Add(self.kpoint_choice, 0, wx.ALIGN_CENTER_VERTICAL | wx.ALL, 5)

        main_sizer.Add(hsz1, 1, wx.EXPAND, 5)

        self.SetSizerAndFit(main_sizer)

    def GetNamelist(self):
        """
        Build and returns a string with whe &SYSTEM Fortran NAMELIST.
        """
        d = OrderedDict(
            ecut=10,
            rprimd="20, 0, 0, 0, 20, 0, 0, 0, 20",
            kpoint=self.kpoint_choice.GetStringSelection()
            #osc_ecut= TODO why here and not in control?
            #nsym=
            #symrel=
        )

        namelist = ["&SYSTEM"]
        for k, v in d.items():
            namelist.append(k + " = " + str(v) + ",")
        namelist.append("/")

        return "\n".join(namelist)


class FFTProfFrame(awx.Frame):
    def __init__(self, parent, **kwargs):
        super(FFTProfFrame, self).__init__(parent, -1, **kwargs)

        panel = wx.Panel(self, -1)

        self.control = ControlPanel(panel)
        self.system = SystemPanel(panel)

        start_button = wx.Button(panel, -1, label='Start Benchmark')
        start_button.Bind(wx.EVT_BUTTON, self.OnStartButton)

        show_button = wx.Button(panel, -1, label='Show Input')
        show_button.Bind(wx.EVT_BUTTON, self.OnShowInput)

        main_sizer = wx.BoxSizer(wx.VERTICAL)
        main_sizer.Add(self.control, 1, wx.EXPAND, 5)
        main_sizer.Add(self.system, 1, wx.EXPAND, 5)

        hsz = wx.BoxSizer(wx.HORIZONTAL)
        hsz.Add(start_button, 0, wx.ALIGN_CENTER_HORIZONTAL, 5)
        hsz.Add(show_button, 0, wx.ALIGN_CENTER_HORIZONTAL, 5)

        main_sizer.Add(hsz, 0, wx.ALIGN_CENTER_HORIZONTAL, 5)

        panel.SetSizerAndFit(main_sizer)

    def MakeInput(self):
        """Build the input file for FFTPROF."""
        fft_input = self.control.GetNamelist() + "\n"
        fft_input += self.system.GetNamelist() + "\n"
        return fft_input

    def OnShowInput(self, event):
        """Show the fftprof input in an external window."""
        fft_input = self.MakeInput()
        SimpleTextViewer(self, fft_input).Show()

    def OnStartButton(self, event):
        """Run the benchmark and plot the results with `matplotlib`."""
        fft_input = self.MakeInput()
        #return

        # Run FFTPROF.
        try:
            fftprof = FFTProf(fft_input)
            returncode = fftprof.run()

            if returncode != 0:
                raise RunTimeError("fftprof exited with returncode %s" % returncode)

            # Analyze the results.
            fftprof.plot()

        except:
            return awx.showErrorMessage(self)


def wxapp_fftprof():
    """Stand alone application."""
    app = awx.App()
    frame = FFTProfFrame(None)
    frame.Show()
    return app


if __name__ == "__main__":
    wxapp_fftprof().MainLoop()



from __future__ import print_function, division

import wx
import inspect
import warnings

from dialogs import showErrorMessage

__all__ = [
    "Func1dPlotFrame",
]


class Func1dPlotFrame(wx.Frame):
    """
    Simple Frame that allows the user to have access to the method
    of a `Function1D` objects and plot the results with wxmplot.
    """

    def __init__(self, parent, func1d, **kwargs):
        """
        Args:
            parent:
                wx parent window.
            func1d:
                `Function1d` object.
        """
        super(Func1dPlotFrame, self).__init__(parent=parent, **kwargs)
        self.func1d = func1d
        self.BuildUi()

        # Store PlotFrames in this list.
        self._pframes = []

    def BuildUi(self):
        panel = wx.Panel(self, id=-1, size=wx.Size(500, 300))

        bSizer10 = wx.BoxSizer(wx.VERTICAL)

        # Get the methods of the objects.
        obj_methods = ["None"]
        obj_methods += [t[0] for t in inspect.getmembers(self.func1d, predicate=inspect.ismethod)
                        if not t[0].startswith("_")]
        sorted(obj_methods)

        self.method_choice = wx.Choice(panel, -1, wx.DefaultPosition, wx.DefaultSize, obj_methods, 0)
        self.method_choice.SetSelection(0)
        bSizer10.Add(self.method_choice, 0, wx.ALL | wx.ALIGN_CENTER_HORIZONTAL | wx.EXPAND, 5)

        fgSizer4 = wx.FlexGridSizer(0, 2, 0, 0)
        fgSizer4.SetFlexibleDirection(wx.BOTH)
        fgSizer4.SetNonFlexibleGrowMode(wx.FLEX_GROWMODE_SPECIFIED)

        self.replot_checkbox = wx.CheckBox(panel, id=-1, label="Replot")
        self.replot_checkbox.SetValue(True)
        fgSizer4.Add(self.replot_checkbox, 0, wx.ALL | wx.ALIGN_CENTER_HORIZONTAL, 5)

        self.plot_button = wx.Button(panel, -1, u"Plot", wx.DefaultPosition, wx.DefaultSize, 0)
        fgSizer4.Add(self.plot_button, 0, wx.ALL | wx.ALIGN_CENTER_HORIZONTAL, 5)

        bSizer10.Add(fgSizer4, 0, 0, 5)

        self.SetSizerAndFit(bSizer10)

        self.plot_button.Bind(wx.EVT_BUTTON, self.OnClick)

    def OnClick(self, event):
        method = self.method_choice.GetStringSelection()

        try:
            from wxmplot import PlotFrame
        except ImportError:
            #warnings.warn("Error while importing wxmplot. Some features won't be available")
            raise

        plotframe = None
        if self.replot_checkbox.GetValue() and len(self._pframes):
            plotframe = self._pframes[-1]

        try:
            if method == "None":
                g = self.func1d
            else:
                g = getattr(self.func1d, method)()

            if plotframe is None:
                plotframe = PlotFrame(self)
                self._pframes.append(plotframe)
                plotframe.plot(g.mesh, g.values, label=method, draw_legend=True)
                plotframe.Show()
            else:
                plotframe.oplot(g.mesh, g.values, label=method, draw_legend=True)

        except:
            showErrorMessage(self)




"""
Pydantic models storing Structure objects and associated metadata

Some of these models are inspired to emmet or we try to field names and models that should be compatible.
"""
from __future__ import annotations

from enum import Enum
from typing import Dict, Any  # List,
from pydantic import Field
from pymatgen.core.composition import Composition
#from pymatgen.core.periodic_table import Element
from abipy.core.structure import Structure
from abipy.htc.base_models import AbipyModel


class CrystalSystem(str, Enum):
    """
    The crystal system of the lattice
    """
    tri = "Triclinic"
    mono = "Monoclinic"
    ortho = "Orthorhombic"
    tet = "Tetragonal"
    trig = "Trigonal"
    hex_ = "Hexagonal"
    cubic = "Cubic"


class StructureData(AbipyModel):
    """
    Store structural info set and symmetry metadata.
    """

    #mpid: str = Field(None, description="MP identifier")
    # Optional or in CustomDict?

    #source_id: str  = Field(..., "Identifier e.g. mp-123")
    #source_type: str = Field(..., "source type e.g. MP for the materials project")
    #meta: Dict[str, Any] = Field(None)

    #custom_data = Dict[str, Any] = Field(None, description="Pseudopotential specifications.")

    #hall:

    crystal_system: CrystalSystem = Field(
        None, title="Crystal System", description="The crystal system for this lattice"
    )

    spg_symbol: str = Field(
        None,
        title="Space Group Symbol",
        description="The spacegroup symbol for the lattice",
    )

    spg_number: int = Field(
        None,
        title="Space Group Number",
        description="The spacegroup number for the lattice",
    )

    point_group: str = Field(
        None, title="Point Group Symbol", description="The point group for the lattice"
    )

    symprec: float = Field(
        None,
        title="Symmetry Finding Precision",
        description="The precision given to spglib to determine the symmetry of this lattice",
    )

    spglib_version: str = Field(None, title="SPGLib version")

    # Structure metadata
    nsites: int = Field(None, description="Total number of sites in the structure")

    #elements: List[Element] = Field(None, description="List of elements in the material")

    nelements: int = Field(None, title="Number of Elements")

    composition: Composition = Field(None, description="Full composition for the material")

    composition_reduced: Composition = Field(
        None,
        title="Reduced Composition",
        description="Simplified representation of the composition",
    )

    formula_pretty: str = Field(
        None,
        title="Pretty Formula",
        description="Cleaned representation of the formula",
    )
    formula_anonymous: str = Field(
        None,
        title="Anonymous Formula",
        description="Anonymized representation of the formula",
    )
    chemsys: str = Field(
        None,
        title="Chemical System",
        description="dash-delimited string of elements in the material",
    )
    volume: float = Field(
        None,
        title="Volume",
        description="Total volume for this structure in Angstroms^3",
    )

    density: float = Field(None, title="Density", description="Density in grams per cm^3")

    density_atomic: float = Field(
        None,
        title="Packing Density",
        description="The atomic packing density in atoms per cm^3",
    )

    structure: Structure = Field(..., description="Abipy Structure object.")

    #custom_dict: Dict[str, Any] = Field(None, description="")

    @classmethod
    def from_mpid(cls, material_id: str) -> StructureData:
        """
        Initialize the model from a MaterialsProject id.
        """
        structure = Structure.from_mpid(material_id, final=True)
        return cls.from_structure(structure)

    @classmethod
    def from_structure(cls, structure: Structure) -> StructureData:
        """
        Initialize the model from a Structure.
        """
        from pymatgen.symmetry.analyzer import SpacegroupAnalyzer, spglib

        #symprec = SETTINGS.SYMPREC
        symprec = 0.1
        sg = SpacegroupAnalyzer(structure, symprec=symprec)
        symmetry: Dict[str, Any] = {"symprec": symprec}
        if not sg.get_symmetry_dataset():
            sg = SpacegroupAnalyzer(structure, 1e-3, 1)
            symmetry["symprec"] = 1e-3

        symmetry.update(
            {
                "spg_symbol": sg.get_space_group_symbol(),
                "spg_number": sg.get_space_group_number(),
                "point_group": sg.get_point_group_symbol(),
                "crystal_system": CrystalSystem(sg.get_crystal_system().title()),
                "hall": sg.get_hall(),
                "spglib_version": spglib.__version__,
            }
        )

        comp = structure.composition.remove_charges()
        elsyms = sorted(set([e.symbol for e in comp.elements]))
        #symmetry = SymmetryData.from_structure(structure)

        data = {
            "nsites": structure.num_sites,
            #"elements": elsyms,
            "nelements": len(elsyms),
            "composition": comp,
            "composition_reduced": comp.reduced_composition,
            "formula_pretty": comp.reduced_formula,
            "formula_anonymous": comp.anonymized_formula,
            "chemsys": "-".join(elsyms),
            "volume": structure.volume,
            "density": structure.density,
            "density_atomic": structure.volume / structure.num_sites,
            #"symmetry": symmetry,
        }

        structure = Structure.as_structure(structure)  #.copy()
        symmetry.update(data)

        return cls(structure=structure, **symmetry)

    def get_title(self) -> str:
        """Return string with metadata. Useful when creating views"""
        return f"Structure: {self.formula_pretty}, {self.spg_symbol} ({self.spg_number}), " + \
               f"{self.crystal_system}, natom: {self.nsites}"

    #def get_panel_view(self):
    #    return



#!/usr/bin/env python
import wx
import random
import numpy as np

from abipy.gui.awx.core import path_img

__all__ = [
    "tetris_game",
    "puzzle_game",
]

# Taken from  http://zetcode.com/wxpython/thetetrisgame/
class Tetris(wx.Frame):
    def __init__(self, parent, id, title):
        wx.Frame.__init__(self, parent, id, title, size=(180, 380))

        self.statusbar = self.CreateStatusBar()
        self.statusbar.SetStatusText('0')
        self.board = Board(self)
        self.board.SetFocus()
        self.board.start()

        self.Centre()
        self.Show(True)
       

class Board(wx.Panel):
    BoardWidth = 10
    BoardHeight = 22
    Speed = 300
    ID_TIMER = 1

    def __init__(self, parent):
        wx.Panel.__init__(self, parent)

        self.timer = wx.Timer(self, Board.ID_TIMER)
        self.isWaitingAfterLine = False
        self.curPiece = Shape()
        self.nextPiece = Shape()
        self.curX = 0
        self.curY = 0
        self.numLinesRemoved = 0
        self.board = []

        self.isStarted = False
        self.isPaused = False

        self.Bind(wx.EVT_PAINT, self.OnPaint)
        self.Bind(wx.EVT_KEY_DOWN, self.OnKeyDown)
        self.Bind(wx.EVT_TIMER, self.OnTimer, id=Board.ID_TIMER)

        self.clearBoard()

    def shapeAt(self, x, y):
        return self.board[(y * Board.BoardWidth) + x]

    def setShapeAt(self, x, y, shape):
        self.board[(y * Board.BoardWidth) + x] = shape

    def squareWidth(self):
        return self.GetClientSize().GetWidth() / Board.BoardWidth

    def squareHeight(self):
        return self.GetClientSize().GetHeight() / Board.BoardHeight

    def start(self):
        if self.isPaused:
            return

        self.isStarted = True
        self.isWaitingAfterLine = False
        self.numLinesRemoved = 0
        self.clearBoard()

        self.newPiece()
        self.timer.Start(Board.Speed)

    def pause(self):
        if not self.isStarted:
            return

        self.isPaused = not self.isPaused
        statusbar = self.GetParent().statusbar

        if self.isPaused:
            self.timer.Stop()
            statusbar.SetStatusText('paused')
        else:
            self.timer.Start(Board.Speed)
            statusbar.SetStatusText(str(self.numLinesRemoved))

        self.Refresh()

    def clearBoard(self):
        for i in range(Board.BoardHeight * Board.BoardWidth):
            self.board.append(Tetrominoes.NoShape)

    def OnPaint(self, event):

        dc = wx.PaintDC(self)        

        size = self.GetClientSize()
        boardTop = size.GetHeight() - Board.BoardHeight * self.squareHeight()
        
        for i in range(Board.BoardHeight):
            for j in range(Board.BoardWidth):
                shape = self.shapeAt(j, Board.BoardHeight - i - 1)
                if shape != Tetrominoes.NoShape:
                    self.drawSquare(dc,
                        0 + j * self.squareWidth(),
                        boardTop + i * self.squareHeight(), shape)

        if self.curPiece.shape() != Tetrominoes.NoShape:
            for i in range(4):
                x = self.curX + self.curPiece.x(i)
                y = self.curY - self.curPiece.y(i)
                self.drawSquare(dc, 0 + x * self.squareWidth(),
                    boardTop + (Board.BoardHeight - y - 1) * self.squareHeight(),
                    self.curPiece.shape())


    def OnKeyDown(self, event):
        if not self.isStarted or self.curPiece.shape() == Tetrominoes.NoShape:
            event.Skip()
            return

        keycode = event.GetKeyCode()

        if keycode in (ord('P'), ord('p')):
            self.pause()
            return
        if self.isPaused:
            return
        elif keycode == wx.WXK_LEFT:
            self.tryMove(self.curPiece, self.curX - 1, self.curY)
        elif keycode == wx.WXK_RIGHT:
            self.tryMove(self.curPiece, self.curX + 1, self.curY)
        elif keycode == wx.WXK_DOWN:
            self.tryMove(self.curPiece.rotatedRight(), self.curX, self.curY)
        elif keycode == wx.WXK_UP:
            self.tryMove(self.curPiece.rotatedLeft(), self.curX, self.curY)
        elif keycode == wx.WXK_SPACE:
            self.dropDown()
        elif keycode == ord('D') or keycode == ord('d'):
            self.oneLineDown()
        else:
            event.Skip()

    def OnTimer(self, event):
        if event.GetId() == Board.ID_TIMER:
            if self.isWaitingAfterLine:
                self.isWaitingAfterLine = False
                self.newPiece()
            else:
                self.oneLineDown()
        else:
            event.Skip()

    def dropDown(self):
        newY = self.curY
        while newY > 0:
            if not self.tryMove(self.curPiece, self.curX, newY - 1):
                break
            newY -= 1

        self.pieceDropped()

    def oneLineDown(self):
        if not self.tryMove(self.curPiece, self.curX, self.curY - 1):
            self.pieceDropped()

    def pieceDropped(self):
        for i in range(4):
            x = self.curX + self.curPiece.x(i)
            y = self.curY - self.curPiece.y(i)
            self.setShapeAt(x, y, self.curPiece.shape())

        self.removeFullLines()

        if not self.isWaitingAfterLine:
            self.newPiece()

    def removeFullLines(self):
        numFullLines = 0

        statusbar = self.GetParent().statusbar

        rowsToRemove = []

        for i in range(Board.BoardHeight):
            n = 0
            for j in range(Board.BoardWidth):
                if not self.shapeAt(j, i) == Tetrominoes.NoShape:
                    n = n + 1

            if n == 10:
                rowsToRemove.append(i)

        rowsToRemove.reverse()

        for m in rowsToRemove:
            for k in range(m, Board.BoardHeight):
                for l in range(Board.BoardWidth):
                        self.setShapeAt(l, k, self.shapeAt(l, k + 1))

            numFullLines = numFullLines + len(rowsToRemove)

            if numFullLines > 0:
                self.numLinesRemoved = self.numLinesRemoved + numFullLines
                statusbar.SetStatusText(str(self.numLinesRemoved)) 
                self.isWaitingAfterLine = True
                self.curPiece.setShape(Tetrominoes.NoShape)
                self.Refresh()

    def newPiece(self):
        self.curPiece = self.nextPiece
        statusbar = self.GetParent().statusbar
        self.nextPiece.setRandomShape()
        self.curX = Board.BoardWidth / 2 + 1
        self.curY = Board.BoardHeight - 1 + self.curPiece.minY()

        if not self.tryMove(self.curPiece, self.curX, self.curY):
            self.curPiece.setShape(Tetrominoes.NoShape)
            self.timer.Stop()
            self.isStarted = False
            statusbar.SetStatusText('Game over')

    def tryMove(self, newPiece, newX, newY):
        for i in range(4):
            x = newX + newPiece.x(i)
            y = newY - newPiece.y(i)
            if x < 0 or x >= Board.BoardWidth or y < 0 or y >= Board.BoardHeight:
                return False
            if self.shapeAt(x, y) != Tetrominoes.NoShape:
                return False

        self.curPiece = newPiece
        self.curX = newX
        self.curY = newY
        self.Refresh()
        return True

    def drawSquare(self, dc, x, y, shape):
        colors = ['#000000', '#CC6666', '#66CC66', '#6666CC',
                  '#CCCC66', '#CC66CC', '#66CCCC', '#DAAA00']

        light = ['#000000', '#F89FAB', '#79FC79', '#7979FC', 
                 '#FCFC79', '#FC79FC', '#79FCFC', '#FCC600']

        dark = ['#000000', '#803C3B', '#3B803B', '#3B3B80', 
                 '#80803B', '#803B80', '#3B8080', '#806200']

        pen = wx.Pen(light[shape])
        pen.SetCap(wx.CAP_PROJECTING)
        dc.SetPen(pen)

        dc.DrawLine(x, y + self.squareHeight() - 1, x, y)
        dc.DrawLine(x, y, x + self.squareWidth() - 1, y)

        darkpen = wx.Pen(dark[shape])
        darkpen.SetCap(wx.CAP_PROJECTING)
        dc.SetPen(darkpen)

        dc.DrawLine(x + 1, y + self.squareHeight() - 1,
            x + self.squareWidth() - 1, y + self.squareHeight() - 1)
        dc.DrawLine(x + self.squareWidth() - 1, 
        y + self.squareHeight() - 1, x + self.squareWidth() - 1, y + 1)

        dc.SetPen(wx.TRANSPARENT_PEN)
        dc.SetBrush(wx.Brush(colors[shape]))
        dc.DrawRectangle(x + 1, y + 1, self.squareWidth() - 2, 
        self.squareHeight() - 2)


class Tetrominoes(object):
    NoShape = 0
    ZShape = 1
    SShape = 2
    LineShape = 3
    TShape = 4
    SquareShape = 5
    LShape = 6
    MirroredLShape = 7


class Shape(object):
    coordsTable = (
        ((0, 0),     (0, 0),     (0, 0),     (0, 0)),
        ((0, -1),    (0, 0),     (-1, 0),    (-1, 1)),
        ((0, -1),    (0, 0),     (1, 0),     (1, 1)),
        ((0, -1),    (0, 0),     (0, 1),     (0, 2)),
        ((-1, 0),    (0, 0),     (1, 0),     (0, 1)),
        ((0, 0),     (1, 0),     (0, 1),     (1, 1)),
        ((-1, -1),   (0, -1),    (0, 0),     (0, 1)),
        ((1, -1),    (0, -1),    (0, 0),     (0, 1))
    )

    def __init__(self):
        self.coords = [[0,0] for i in range(4)]
        self.pieceShape = Tetrominoes.NoShape

        self.setShape(Tetrominoes.NoShape)

    def shape(self):
        return self.pieceShape

    def setShape(self, shape):
        table = Shape.coordsTable[shape]
        for i in range(4):
            for j in range(2):
                self.coords[i][j] = table[i][j]

        self.pieceShape = shape

    def setRandomShape(self):
        self.setShape(random.randint(1, 7))

    def x(self, index):
        return self.coords[index][0]

    def y(self, index):
        return self.coords[index][1]

    def setX(self, index, x):
        self.coords[index][0] = x

    def setY(self, index, y):
        self.coords[index][1] = y

    def minX(self):
        m = self.coords[0][0]
        for i in range(4):
            m = min(m, self.coords[i][0])

        return m

    def maxX(self):
        m = self.coords[0][0]
        for i in range(4):
            m = max(m, self.coords[i][0])

        return m

    def minY(self):
        m = self.coords[0][1]
        for i in range(4):
            m = min(m, self.coords[i][1])

        return m

    def maxY(self):
        m = self.coords[0][1]
        for i in range(4):
            m = max(m, self.coords[i][1])

        return m

    def rotatedLeft(self):
        if self.pieceShape == Tetrominoes.SquareShape:
            return self

        result = Shape()
        result.pieceShape = self.pieceShape
        for i in range(4):
            result.setX(i, self.y(i))
            result.setY(i, -self.x(i))

        return result

    def rotatedRight(self):
        if self.pieceShape == Tetrominoes.SquareShape:
            return self

        result = Shape()
        result.pieceShape = self.pieceShape
        for i in range(4):
            result.setX(i, -self.y(i))
            result.setY(i, self.x(i))

        return result

def tetris_game():
    app = wx.App()
    Tetris(None, -1, 'Tetris')
    app.MainLoop()


class Puzzle(wx.Dialog):
    
    def __init__(self, *args, **kw):
        super(Puzzle, self).__init__(*args, **kw)
        
        self.InitUI()
        
    def InitUI(self):
        # convert image.png -crop 120x90 image%d.png
        # image should be 360x270
        images = ["abipy" + str(i) + ".png" for i in range(9)]
        images =  map(path_img, images)

        self.pos = np.reshape(list(range(9)), (3,3))

        self.sizer = wx.GridSizer(3, 3, 0, 0)

        numbers = list(range(8))
        random.shuffle(numbers)

        for i in numbers:
            btn = wx.BitmapButton(self, i, wx.Bitmap(images[i]))
            btn.Bind(wx.EVT_BUTTON, self.OnPressButton, btn)
            self.sizer.Add(btn)

        self.empty = wx.BitmapButton(self, bitmap=wx.Bitmap(path_img('empty.png')))
        self.empty.Bind(wx.EVT_BUTTON, self.OnPressButton, self.empty)
        self.sizer.Add(self.empty)

        self.SetSizerAndFit(self.sizer)
        self.SetTitle('Puzzle')
        self.Centre()
        self.ShowModal()
        self.Destroy()

    def OnPressButton(self, e):
        
        btn = e.GetEventObject()
        
        width = self.empty.GetSize().x
        height = self.empty.GetSize().y

        btnX = btn.GetPosition().x
        btnY = btn.GetPosition().y
        emptyX = self.empty.GetPosition().x
        emptyY = self.empty.GetPosition().y
    
        
        if (((btnX == emptyX) and (emptyY - btnY) == height)
          or ((btnX == emptyX) and (emptyY - btnY) == -height)
          or ((btnY == emptyY) and (emptyX - btnX) == width)
          or ((btnY == emptyY) and (emptyX - btnX) == -width)):
                 
            self.ExchangeImages(btn)
            
    def ExchangeImages(self, btn):
        bmp1 = self.empty.GetBitmapLabel()
        bmp2 = btn.GetBitmapLabel()
        
        self.empty.SetBitmapLabel(bmp2)
        btn.SetBitmapLabel(bmp1)
        
        self.empty = btn        


def puzzle_game():
    ex = wx.App()
    Puzzle(None)
    ex.MainLoop()    


if __name__ == "__main__":
    #tetris_game()
    puzzle_game()



"""Panels for interacting with output files in text format."""

import panel as pn
import panel.widgets as pnw
import bokeh.models.widgets as bkw

from abipy.panels.core import AbipyParameterized, Loading, mpl, ply, dfc


class AbinitOutputFilePanel(AbipyParameterized):
    """
    Panel with widgets to interact with the main Abinit output file.
    """
    def __init__(self, outfile, **params):
        super().__init__(**params)
        self.outfile = outfile

    def get_cycles_view(self, what):
        """
        Return GridBox with matplotlib plots for the GS/DFPT SCF cycles
        or None if no cycle is found.
        """
        if what == "GS_SCF":
            cycles = self.outfile.get_all_gs_scf_cycles()
        elif what == "DFPT_D2DE":
            cycles = self.outfile.get_all_d2de_scf_cycles()
        else:
            raise ValueError(f"Invalid value for what: `{what}`")

        if not cycles: return None

        num_plots, nrows, ncols = len(cycles), 1, 1
        if num_plots > 1:
            ncols = 2
            nrows = (num_plots // ncols) + (num_plots % ncols)

        box = pn.GridBox(nrows=nrows, ncols=ncols, sizing_mode='stretch_width')
        #box = pn.Column(sizing_mode="stretch_width")

        for icycle, cycle in enumerate(cycles):
            #box.append(mpl(cycle.plot(title="%s cycle #%d" % (what, icycle), **self.mpl_kwargs)))
            f = ply(cycle.plotly(title="%s cycle #%d" % (what, icycle + 1), show=False))
            box.append(f)

        return box

    def get_outfile_view(self):
        col = pn.Column(sizing_mode="stretch_width")
        ca = col.append; cext = col.extend

        filepath = self.outfile.filepath
        with open(filepath, "rt") as fh:
            text = fh.read()

        ace = pnw.Ace(value=text, language='text', readonly=True,
                      sizing_mode='stretch_width', height=1200)
                      #sizing_mode='stretch_width', width=900)
        cext([f"## Output <small>{filepath}</small>", ace, pn.layout.Divider()])

        return col

    def get_panel(self, as_dict=False, **kwargs):
        """Return tabs with widgets to interact with the Abinit output file."""
        d = {}

        d["Summary"] = self.get_summary_view_for_abiobj(self.outfile)
        d["Output"] = self.get_outfile_view()

        df = self.outfile.get_dims_spginfo_dataframe()
        df.index.name = "Dataset"
        d["Dims"] = dfc(df, transpose=True)

        # Add tabs with plots for the GS/DFPT SCF cycles.
        for what in ("GS_SCF", "DFPT_D2DE"):
            box = self.get_cycles_view(what)
            if box is not None:
                d[f"{what} cycles"] = box

        #timer = self.get_timer()
        #timer.plot_all(**self.mpl_kwargs)

        if as_dict: return d

        return self.get_template_from_tabs(d, template=kwargs.get("template", None))


class AbinitOutputFilePanelWithFileInput(AbipyParameterized):

    info_str = """
This application allows users to analyze the Abinit main output file
"""

    def __init__(self, **params):
        super().__init__(**params)

        help_md = pn.pane.Markdown(f"""
## Description

{self.info_str}
""")

        self.main_area = pn.Column(help_md, sizing_mode="stretch_width")
        self.abifile = None

        self.file_input = pnw.FileInput(height=60, css_classes=["pnx-file-upload-area"])
        self.file_input.param.watch(self.on_file_input, "value")

    def on_file_input(self, event):

        with Loading(self.main_area):
            new_abifile = self.get_abifile_from_file_input(self.file_input)

            if self.abifile is not None:
                self.abifile.remove()

            self.abifile = new_abifile
            self.main_area.objects = [self.abifile.get_panel()]

    def get_panel(self, **kwargs):

        col = pn.Column("## Upload (or drag & drop) an *.abo* file (main ABINIT output file):",
                        self.get_fileinput_section(self.file_input),
                        sizing_mode="stretch_width")

        main = pn.Column(col, self.main_area, sizing_mode="stretch_width")

        cls, cls_kwds = self.get_abinit_template_cls_kwds()

        return cls(main=main, title="Abo Analyzer", **cls_kwds)



# coding: utf-8
"""
Utilities for generating matplotlib plots.

.. note::

    Avoid importing matplotlib or plotly in the module namespace otherwise startup is very slow.
"""
from __future__ import annotations

import os
import time
import itertools
import numpy as np

from collections import OrderedDict, namedtuple
from typing import Any, List
from pymatgen.util.plotting import add_fig_kwargs, get_ax_fig_plt, get_ax3d_fig_plt, get_axarray_fig_plt
from .numtools import data_from_cplx_mode


__all__ = [
    "set_axlims",
    "add_fig_kwargs",
    "get_ax_fig_plt",
    "get_ax3d_fig_plt",
    "plot_array",
    "ArrayPlotter",
    "data_from_cplx_mode",
    "Marker",
    "plot_unit_cell",
    "GenericDataFilePlotter",
    "GenericDataFilesPlotter",
    "add_plotly_fig_kwargs",
    "get_fig_plotly",
    "get_figs_plotly",
]


# https://matplotlib.org/gallery/lines_bars_and_markers/linestyles.html
linestyles = OrderedDict(
    [('solid',               (0, ())),
     ('loosely_dotted',      (0, (1, 10))),
     ('dotted',              (0, (1, 5))),
     ('densely_dotted',      (0, (1, 1))),

     ('loosely_dashed',      (0, (5, 10))),
     ('dashed',              (0, (5, 5))),
     ('densely_dashed',      (0, (5, 1))),

     ('loosely_dashdotted',  (0, (3, 10, 1, 10))),
     ('dashdotted',          (0, (3, 5, 1, 5))),
     ('densely_dashdotted',  (0, (3, 1, 1, 1))),

     ('loosely_dashdotdotted', (0, (3, 10, 1, 10, 1, 10))),
     ('dashdotdotted',         (0, (3, 5, 1, 5, 1, 5))),
     ('densely_dashdotdotted', (0, (3, 1, 1, 1, 1, 1)))]
)


###################
# Matplotlib tools
###################

def is_mpl_figure(obj: Any) -> bool:
    """Return True if obj is a matplotlib Figure."""
    from matplotlib import pyplot as plt
    return isinstance(obj, plt.Figure)


def ax_append_title(ax, title, loc="center", fontsize=None) -> str:
    """Add title to previous ax.title. Return new title."""
    prev_title = ax.get_title(loc=loc)
    new_title = prev_title + title
    ax.set_title(new_title, loc=loc, fontsize=fontsize)
    return new_title


def ax_share(xy_string: str, *ax_list) -> None:
    """
    Share x- or y-axis of two or more subplots after they are created

    Args:
        xy_string: "x" to share x-axis, "xy" for both
        ax_list: List of axes to share.

    Example:

        ax_share("y", ax0, ax1)
        ax_share("xy", *(ax0, ax1, ax2))
    """
    if "x" in xy_string:
        for ix, ax in enumerate(ax_list):
            others = [a for a in ax_list if a != ax]
            ax.get_shared_x_axes().join(*others)

    if "y" in xy_string:
        for ix, ax in enumerate(ax_list):
            others = [a for a in ax_list if a != ax]
            ax.get_shared_y_axes().join(*others)


def set_axlims(ax, lims, axname: str) -> tuple:
    """
    Set the data limits for the axis ax.

    Args:
        lims: tuple(2) for (left, right), tuple(1) or scalar for left only.
        axname: "x" for x-axis, "y" for y-axis.

    Return: (left, right)
    """
    left, right = None, None
    if lims is None: return left, right

    len_lims = None
    try:
        len_lims = len(lims)
    except TypeError:
        # Assume Scalar
        left = float(lims)

    if len_lims is not None:
        if len(lims) == 2:
            left, right = lims[0], lims[1]
        elif len(lims) == 1:
            left = lims[0]

    set_lim = getattr(ax, {"x": "set_xlim", "y": "set_ylim"}[axname])
    if left != right:
        set_lim(left, right)

    return left, right


def set_ax_xylabels(ax, xlabel: str, ylabel: str, exchange_xy: bool) -> None:
    """
    Set the x- and the y-label of axis ax, exchanging x and y if exchange_xy
    """
    if exchange_xy: xlabel, ylabel = ylabel, xlabel
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)


def set_visible(ax, boolean: bool, *args) -> None:
    """
    Hide/Show the artists of axis ax listed in args.
    """
    if "legend" in args and ax.legend():
        ax.legend().set_visible(boolean)
    if "title" in args and ax.title:
        ax.title.set_visible(boolean)
    if "xlabel" in args and ax.xaxis.label:
        ax.xaxis.label.set_visible(boolean)
    if "ylabel" in args and ax.yaxis.label:
        ax.yaxis.label.set_visible(boolean)
    if "xticklabels" in args:
        for label in ax.get_xticklabels():
            label.set_visible(boolean)
    if "yticklabels" in args:
        for label in ax.get_yticklabels():
            label.set_visible(boolean)


def rotate_ticklabels(ax, rotation: float, axname: str ="x") -> None:
    """Rotate the ticklables of axis ``ax``"""
    if "x" in axname:
        for tick in ax.get_xticklabels():
            tick.set_rotation(rotation)
    if "y" in axname:
        for tick in ax.get_yticklabels():
            tick.set_rotation(rotation)


@add_fig_kwargs
def plot_xy_with_hue(data, x, y, hue, decimals=None, ax=None,
                     xlims=None, ylims=None, fontsize=12, **kwargs):
    """
    Plot y = f(x) relation for different values of `hue`.
    Useful for convergence tests done wrt to two parameters.

    Args:
        data: |pandas-DataFrame| containing columns `x`, `y`, and `hue`.
        x: Name of the column used as x-value
        y: Name of the column(s) used as y-value
        hue: Variable that define subsets of the data, which will be drawn on separate lines
        decimals: Number of decimal places to round `hue` columns. Ignore if None
        ax: |matplotlib-Axes| or None if a new figure should be created.
        xlims, ylims: Set the data limits for the x(y)-axis. Accept tuple e.g. `(left, right)`
            or scalar e.g. `left`. If left (right) is None, default values are used
        fontsize: Legend fontsize.
        kwargs: Keywork arguments are passed to ax.plot method.

    Returns: |matplotlib-Figure|
    """
    if isinstance(y, (list, tuple)):
        # Recursive call for each ax in ax_list.
        num_plots, ncols, nrows = len(y), 1, 1
        if num_plots > 1:
            ncols = 2
            nrows = (num_plots // ncols) + (num_plots % ncols)

        ax_list, fig, plt = get_axarray_fig_plt(None, nrows=nrows, ncols=ncols,
                                                sharex=False, sharey=False, squeeze=False)

        ax_list = ax_list.ravel()
        if num_plots % ncols != 0: ax_list[-1].axis('off')

        for yname, ax in zip(y, ax_list):
            plot_xy_with_hue(data, x, str(yname), hue, decimals=decimals, ax=ax,
                             xlims=xlims, ylims=ylims, fontsize=fontsize, show=False, **kwargs)
        return fig

    # Check here because pandas error messages are a bit criptic.
    miss = [k for k in (x, y, hue) if k not in data]
    if miss:
        raise ValueError("Cannot find `%s` in dataframe.\nAvailable keys are: %s" % (str(miss), str(data.keys())))

    # Truncate values in hue column so that we can group.
    if decimals is not None:
        data = data.round({hue: decimals})

    ax, fig, plt = get_ax_fig_plt(ax=ax)
    for key, grp in data.groupby(hue):
        # Sort xs and rearrange ys
        xy = np.array(sorted(zip(grp[x], grp[y]), key=lambda t: t[0]))
        xvals, yvals = xy[:, 0], xy[:, 1]

        #label = "{} = {}".format(hue, key)
        label = "%s" % (str(key))
        if not kwargs:
            ax.plot(xvals, yvals, 'o-', label=label)
        else:
            ax.plot(xvals, yvals, label=label, **kwargs)

    ax.grid(True)
    ax.set_xlabel(x)
    ax.set_ylabel(y)
    set_axlims(ax, xlims, "x")
    set_axlims(ax, ylims, "y")
    ax.legend(loc="best", fontsize=fontsize, shadow=True)

    return fig


@add_fig_kwargs
def plot_array(array, color_map=None, cplx_mode="abs", **kwargs):
    """
    Use imshow for plotting 2D or 1D arrays.

    Example::

        plot_array(np.random.rand(10,10))

    See <http://stackoverflow.com/questions/7229971/2d-grid-data-visualization-in-python>

    Args:
        array: Array-like object (1D or 2D).
        color_map: color map.
        cplx_mode:
            Flag defining how to handle complex arrays. Possible values in ("re", "im", "abs", "angle")
            "re" for the real part, "im" for the imaginary part.
            "abs" means that the absolute value of the complex number is shown.
            "angle" will display the phase of the complex number in radians.

    Returns: |matplotlib-Figure|
    """
    # Handle vectors
    array = np.atleast_2d(array)
    array = data_from_cplx_mode(cplx_mode, array)

    import matplotlib as mpl
    from matplotlib import pyplot as plt
    if color_map is None:
        # make a color map of fixed colors
        color_map = mpl.colors.LinearSegmentedColormap.from_list('my_colormap',
                                                                 ['blue', 'black', 'red'], 256)

    img = plt.imshow(array, interpolation='nearest', cmap=color_map, origin='lower')

    # Make a color bar
    plt.colorbar(img, cmap=color_map)

    # Set grid
    plt.grid(True, color='white')

    fig = plt.gcf()
    return fig


class ArrayPlotter:

    def __init__(self, *labels_and_arrays):
        """
        Args:
            labels_and_arrays: List [("label1", arr1), ("label2", arr2")]
        """
        self._arr_dict = OrderedDict()
        for label, array in labels_and_arrays:
            self.add_array(label, array)

    def __len__(self):
        return len(self._arr_dict)

    def __iter__(self):
        return self._arr_dict.__iter__()

    def keys(self):
        return self._arr_dict.keys()

    def items(self):
        return self._arr_dict.items()

    def add_array(self, label, array):
        """Add array with the given name."""
        if label in self._arr_dict:
            raise ValueError("%s is already in %s" % (label, list(self._arr_dict.keys())))

        self._arr_dict[label] = array

    def add_arrays(self, labels, arr_list):
        """
        Add a list of arrays

        Args:
            labels: List of labels.
            arr_list: List of arrays.
        """
        assert len(labels) == len(arr_list)
        for label, arr in zip(labels, arr_list):
            self.add_array(label, arr)

    @add_fig_kwargs
    def plot(self, cplx_mode="abs", colormap="jet", fontsize=8, **kwargs):
        """
        Args:
            cplx_mode: "abs" for absolute value, "re", "im", "angle"
            colormap: matplotlib colormap.
            fontsize: legend and label fontsize.

        Returns: |matplotlib-Figure|
        """
        # Build grid of plots.
        num_plots, ncols, nrows = len(self), 1, 1
        if num_plots > 1:
            ncols = 2
            nrows = num_plots // ncols + (num_plots % ncols)

        import matplotlib.pyplot as plt
        fig, ax_mat = plt.subplots(nrows=nrows, ncols=ncols, sharex=False, sharey=False, squeeze=False)
        # Don't show the last ax if num_plots is odd.
        if num_plots % ncols != 0: ax_mat[-1, -1].axis("off")

        from mpl_toolkits.axes_grid1 import make_axes_locatable
        from matplotlib.ticker import MultipleLocator

        for ax, (label, arr) in zip(ax_mat.flat, self.items()):
            data = data_from_cplx_mode(cplx_mode, arr)
            # Use origin to place the [0, 0] index of the array in the lower left corner of the axes.
            img = ax.matshow(data, interpolation='nearest', cmap=colormap, origin='lower', aspect="auto")
            ax.set_title("(%s) %s" % (cplx_mode, label), fontsize=fontsize)

            # Make a color bar for this ax
            # Create divider for existing axes instance
            # http://stackoverflow.com/questions/18266642/multiple-imshow-subplots-each-with-colorbar
            divider3 = make_axes_locatable(ax)
            # Append axes to the right of ax, with 10% width of ax
            cax3 = divider3.append_axes("right", size="10%", pad=0.05)
            # Create colorbar in the appended axes
            # Tick locations can be set with the kwarg `ticks`
            # and the format of the ticklabels with kwarg `format`
            cbar3 = plt.colorbar(img, cax=cax3, ticks=MultipleLocator(0.2), format="%.2f")
            # Remove xticks from ax
            ax.xaxis.set_visible(False)
            # Manually set ticklocations
            #ax.set_yticks([0.0, 2.5, 3.14, 4.0, 5.2, 7.0])

            # Set grid
            ax.grid(True, color='white')

        fig.tight_layout()
        return fig


#TODO use object and introduce c for color, client code should be able to customize it.
# Rename it to ScatterData
class Marker(namedtuple("Marker", "x y s")):
    """
    Stores the position and the size of the marker.
    A marker is a list of tuple(x, y, s) where x, and y are the position
    in the graph and s is the size of the marker.
    Used for plotting purpose e.g. QP data, energy derivatives...

    Example::

        x, y, s = [1, 2, 3], [4, 5, 6], [0.1, 0.2, -0.3]
        marker = Marker(x, y, s)
        marker.extend((x, y, s))

    """
    def __new__(cls, *xys):
        """Extends the base class adding consistency check."""
        if not xys:
            xys = ([], [], [])
            return super().__new__(cls, *xys)

        if len(xys) != 3:
            raise TypeError("Expecting 3 entries in xys got %d" % len(xys))

        x = np.asarray(xys[0])
        y = np.asarray(xys[1])
        s = np.asarray(xys[2])
        xys = (x, y, s)

        for s in xys[-1]:
            if np.iscomplex(s):
                raise ValueError("Found ambiguous complex entry %s" % str(s))

        return super().__new__(cls, *xys)

    def __bool__(self):
        return bool(len(self.s))

    __nonzero__ = __bool__

    def extend(self, xys):
        """
        Extend the marker values.
        """
        if len(xys) != 3:
            raise TypeError("Expecting 3 entries in xys got %d" % len(xys))

        self.x.extend(xys[0])
        self.y.extend(xys[1])
        self.s.extend(xys[2])

        lens = np.array((len(self.x), len(self.y), len(self.s)))
        if np.any(lens != lens[0]):
            raise TypeError("x, y, s vectors should have same lengths but got %s" % str(lens))

    def posneg_marker(self):
        """
        Split data into two sets: the first one contains all the points with positive size.
        The first set contains all the points with negative size.
        """
        pos_x, pos_y, pos_s = [], [], []
        neg_x, neg_y, neg_s = [], [], []

        for x, y, s in zip(self.x, self.y, self.s):
            if s >= 0.0:
                pos_x.append(x)
                pos_y.append(y)
                pos_s.append(s)
            else:
                neg_x.append(x)
                neg_y.append(y)
                neg_s.append(s)

        return self.__class__(pos_x, pos_y, pos_s), Marker(neg_x, neg_y, neg_s)


class MplExpose: # pragma: no cover
    """
    Context manager used to produce several matplotlib figures and then show
    all them at the end so that the user does not need to close the window to
    visualize to the next one.

    Example:

        with MplExpose() as e:
            e(obj.plot1(show=False))
            e(obj.plot2(show=False))
    """
    def __init__(self, slide_mode=False, slide_timeout=None, verbose=1):
        """
        Args:
            slide_mode: If Rrue, iterate over figures. Default: Expose all figures at once.
            slide_timeout: Close figure after slide-timeout seconds. Block if None.
            verbose: verbosity level
        """
        self.figures = []
        self.slide_mode = bool(slide_mode)
        self.timeout_ms = slide_timeout
        self.verbose = verbose
        if self.timeout_ms is not None:
            self.timeout_ms = int(self.timeout_ms * 1000)
            assert self.timeout_ms >= 0

        if self.verbose:
            if self.slide_mode:
                print("\nSliding matplotlib figures with slide timeout: %s [s]" % slide_timeout)
            else:
                print("\nLoading all matplotlib figures before showing them. It may take some time...")

        self.start_time = time.time()

    def __call__(self, obj):
        """
        Add an object to MplExpose.
        Support mpl figure, list of figures or generator yielding figures.
        """
        import types
        if isinstance(obj, (types.GeneratorType, list, tuple)):
            for fig in obj:
                self.add_fig(fig)
        else:
            self.add_fig(obj)

    def add_fig(self, fig):
        """Add a matplotlib figure."""
        if fig is None: return

        if not self.slide_mode:
            self.figures.append(fig)
        else:
            #print("Printing and closing", fig)
            import matplotlib.pyplot as plt
            if self.timeout_ms is not None:
                # Creating a timer object
                # timer calls plt.close after interval milliseconds to close the window.
                timer = fig.canvas.new_timer(interval=self.timeout_ms)
                timer.add_callback(plt.close, fig)
                timer.start()

            plt.show()
            fig.clear()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Activated at the end of the with statement. """
        if exc_type is not None: return
        self.expose()

    def expose(self):
        """Show all figures. Clear figures if needed."""
        if not self.slide_mode:
            print("All figures in memory, elapsed time: %.3f s" % (time.time() - self.start_time))
            import matplotlib.pyplot as plt
            plt.show()
            for fig in self.figures:
                fig.clear()


class PanelExpose:  # pragma: no cover
    """
    Context manager used to produce several matplotlib/plotly figures and then show
    all them inside the Browser using a panel template.

    Example:

        with PanelExpose() as e:
            e(obj.plot1(show=False))
            e(obj.plot2(show=False))
    """
    def __init__(self, title=None, verbose=1):
        """
        Args:
            title: String to be show in the header.
            verbose: verbosity level
        """
        self.title = title
        self.figures = []
        self.verbose = verbose

        if self.verbose:
            print("\nLoading all figures before showing them. It may take some time...")

        self.start_time = time.time()

    def __call__(self, obj):
        """
        Add an object to MplPanelExpose.
        Support mpl figure, list of figures or generator yielding figures.
        """
        import types
        if isinstance(obj, (types.GeneratorType, list, tuple)):
            for fig in obj:
                self.add_fig(fig)
        else:
            self.add_fig(obj)

    def add_fig(self, fig):
        """Add a matplotlib figure."""
        if fig is None: return

        self.figures.append(fig)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Activated at the end of the with statement. """
        if exc_type is not None: return
        self.expose()

    def expose(self):
        """Show all figures. Clear figures if needed."""
        import panel as pn
        pn.config.sizing_mode = 'stretch_width'
        from abipy.panels.core import get_template_cls_from_name
        cls = get_template_cls_from_name("FastGridTemplate")

        template = cls(
            title=self.title if self.title is not None else self.__class__.__name__,
            header_background="#ff8c00 ", # Dark orange
        )
        #pn.config.sizing_mode = 'stretch_width'
        from abipy.panels.core import mpl, ply
        for i, fig in enumerate(self.figures):
            row, col = divmod(i, 2)
            if is_plotly_figure(fig):
                p = ply(fig, with_divider=False)
            elif is_mpl_figure(fig):
                p = mpl(fig, with_divider=False)
            else:
                raise TypeError(f"Don't know how to handle type: `{type(fig)}`")

            if hasattr(template.main, "append"):
                template.main.append(p)
            else:
                # Assume .main area acts like a GridSpec
                row_slice = slice(3 * row, 3 * (row + 1))
                if col == 0: template.main[row_slice, :6] = p
                if col == 1: template.main[row_slice, 6:] = p

        return template.show()


def plot_unit_cell(lattice, ax=None, **kwargs):
    """
    Adds the unit cell of the lattice to a matplotlib Axes3D

    Args:
        lattice: Lattice object
        ax: matplotlib :class:`Axes3D` or None if a new figure should be created.
        kwargs: kwargs passed to the matplotlib function 'plot'. Color defaults to black
            and linewidth to 3.

    Returns:
        matplotlib figure and ax
    """
    ax, fig, plt = get_ax3d_fig_plt(ax)

    if "color" not in kwargs: kwargs["color"] = "k"
    if "linewidth" not in kwargs: kwargs["linewidth"] = 3

    v = 8 * [None]
    v[0] = lattice.get_cartesian_coords([0.0, 0.0, 0.0])
    v[1] = lattice.get_cartesian_coords([1.0, 0.0, 0.0])
    v[2] = lattice.get_cartesian_coords([1.0, 1.0, 0.0])
    v[3] = lattice.get_cartesian_coords([0.0, 1.0, 0.0])
    v[4] = lattice.get_cartesian_coords([0.0, 1.0, 1.0])
    v[5] = lattice.get_cartesian_coords([1.0, 1.0, 1.0])
    v[6] = lattice.get_cartesian_coords([1.0, 0.0, 1.0])
    v[7] = lattice.get_cartesian_coords([0.0, 0.0, 1.0])

    for i, j in ((0, 1), (1, 2), (2, 3), (0, 3), (3, 4), (4, 5), (5, 6),
                 (6, 7), (7, 4), (0, 7), (1, 6), (2, 5), (3, 4)):
        ax.plot(*zip(v[i], v[j]), **kwargs)

    # Plot cartesian frame
    ax_add_cartesian_frame(ax)

    return fig, ax


def ax_add_cartesian_frame(ax, start=(0, 0, 0)):
    """
    Add cartesian frame to 3d axis at point `start`.
    """
    # https://stackoverflow.com/questions/22867620/putting-arrowheads-on-vectors-in-matplotlibs-3d-plot
    from matplotlib.patches import FancyArrowPatch
    from mpl_toolkits.mplot3d import proj3d
    arrow_opts = {"color": "k"}
    arrow_opts.update(dict(lw=1, arrowstyle="-|>",))

    class Arrow3D(FancyArrowPatch):
        def __init__(self, xs, ys, zs, *args, **kwargs):
            FancyArrowPatch.__init__(self, (0, 0), (0, 0), *args, **kwargs)
            self._verts3d = xs, ys, zs

        def draw(self, renderer):
            xs3d, ys3d, zs3d = self._verts3d
            xs, ys, zs = proj3d.proj_transform(xs3d, ys3d, zs3d, renderer.M)
            self.set_positions((xs[0], ys[0]), (xs[1], ys[1]))
            FancyArrowPatch.draw(self, renderer)

    start = np.array(start)
    for end in ((1, 0, 0), (0, 1, 0), (0, 0, 1)):
        end = start + np.array(end)
        xs, ys, zs = list(zip(start, end))
        p = Arrow3D(xs, ys, zs,
                   connectionstyle='arc3', mutation_scale=20,
                   alpha=0.8, **arrow_opts)
        ax.add_artist(p)

    return ax


def plot_structure(structure, ax=None, to_unit_cell=False, alpha=0.7,
                   style="points+labels", color_scheme="VESTA", **kwargs):
    """
    Plot structure with matplotlib (minimalistic version).

    Args:
        structure: |Structure| object
        ax: matplotlib :class:`Axes3D` or None if a new figure should be created.
        alpha: The alpha blending value, between 0 (transparent) and 1 (opaque)
        to_unit_cell: True if sites should be wrapped into the first unit cell.
        style: "points+labels" to show atoms sites with labels.
        color_scheme: color scheme for atom types. Allowed values in ("Jmol", "VESTA")

    Returns: |matplotlib-Figure|
    """
    fig, ax = plot_unit_cell(structure.lattice, ax=ax, linewidth=1)

    from pymatgen.analysis.molecule_structure_comparator import CovalentRadius
    from pymatgen.vis.structure_vtk import EL_COLORS
    xyzs, colors = np.empty((len(structure), 4)), []

    for i, site in enumerate(structure):
        symbol = site.specie.symbol
        color = tuple(i / 255 for i in EL_COLORS[color_scheme][symbol])
        radius = CovalentRadius.radius[symbol]
        if to_unit_cell and hasattr(site, "to_unit_cell"): site = site.to_unit_cell()
        # Use cartesian coordinates.
        x, y, z = site.coords
        xyzs[i] = (x, y, z, radius)
        colors.append(color)
        if "labels" in style:
            ax.text(x, y, z, symbol)

    # The definition of sizes is not optimal because matplotlib uses points
    # wherease we would like something that depends on the radius (5000 seems to give reasonable plots)
    # For possibile approaches, see
    # https://stackoverflow.com/questions/9081553/python-scatter-plot-size-and-style-of-the-marker/24567352#24567352
    # https://gist.github.com/syrte/592a062c562cd2a98a83
    if "points" in style:
        x, y, z, s = xyzs.T.copy()
        s = 5000 * s ** 2
        ax.scatter(x, y, zs=z, s=s, c=colors, alpha=alpha)  #facecolors="white", #edgecolors="blue"

    ax.set_title(structure.composition.formula)
    ax.set_axis_off()

    return fig


def _generic_parser_fh(fh):
    """
    Parse file with data in tabular format. Supports multi datasets a la gnuplot.
    Mainly used for files without any schema, not even CSV

    Args:
        fh: File object

    Returns:
        OrderedDict title --> numpy array
        where title is taken from the first (non-empty) line preceding the dataset
    """
    arr_list = [None]
    data = []
    head_list = []
    count = -1
    last_header = None
    for l in fh:
        l = l.strip()
        if not l or l.startswith("#"):
            count = -1
            last_header = l
            if arr_list[-1] is not None: arr_list.append(None)
            continue

        count += 1
        if count == 0: head_list.append(last_header)
        if arr_list[-1] is None: arr_list[-1] = []
        data = arr_list[-1]
        data.append(list(map(float, l.split())))

    if len(head_list) != len(arr_list):
        raise RuntimeError("len(head_list) != len(arr_list), %d != %d" % (len(head_list), len(arr_list)))

    od = OrderedDict()
    for key, data in zip(head_list, arr_list):
        key = " ".join(key.split())
        if key in od:
            print("Header %s already in dictionary. Using new key %s" % (key, 2 * key))
            key = 2 * key
        od[key] = np.array(data).T.copy()

    return od


class GenericDataFilePlotter(object):
    """
    Extract data from a generic text file with results
    in tabular format and plot data with matplotlib.
    Multiple datasets are supported.
    No attempt is made to handle metadata (e.g. column name)
    Mainly used to handle text files written without any schema.
    """
    def __init__(self, filepath):
        with open(filepath, "rt") as fh:
            self.od = _generic_parser_fh(fh)

    def __str__(self):
        return self.to_string()

    def to_string(self, verbose: int = 0) -> str:
        """String representation with verbosity level `verbose`."""
        lines = []
        for key, arr in self.od.items():
            lines.append("key: `%s` --> array shape: %s" % (key, str(arr.shape)))
        return "\n".join(lines)

    @add_fig_kwargs
    def plot(self, use_index=False, fontsize=8, **kwargs):
        """
        Plot all arrays. Use multiple axes if datasets.

        Args:
            use_index: By default, the x-values are taken from the first column.
                If use_index is False, the x-values are the row index.
            fontsize: fontsize for title.
            kwargs: options passed to ``ax.plot``.

        Return: |matplotlib-figure|
        """
        # build grid of plots.
        num_plots, ncols, nrows = len(self.od), 1, 1
        if num_plots > 1:
            ncols = 2
            nrows = (num_plots // ncols) + (num_plots % ncols)

        ax_list, fig, plt = get_axarray_fig_plt(None, nrows=nrows, ncols=ncols,
                                                sharex=False, sharey=False, squeeze=False)
        ax_list = ax_list.ravel()

        # Don't show the last ax if num_plots is odd.
        if num_plots % ncols != 0: ax_list[-1].axis("off")

        for ax, (key, arr) in zip(ax_list, self.od.items()):
            ax.set_title(key, fontsize=fontsize)
            ax.grid(True)
            xs = arr[0] if not use_index else list(range(len(arr[0])))
            for ys in arr[1:] if not use_index else arr:
                ax.plot(xs, ys)

        return fig


class GenericDataFilesPlotter:

    @classmethod
    def from_files(cls, filepaths: List[str]) -> GenericDataFilesPlotter:
        """
        Build object from a list of `filenames`.
        """
        new = cls()
        for filepath in filepaths:
            new.add_file(filepath)
        return new

    def __init__(self):
        self.odlist = []
        self.filepaths = []

    def __str__(self):
        return self.to_string()

    def to_string(self, verbose: int = 0) -> str:
        lines = []
        app = lines.append
        for od, filepath in zip(self.odlist, self.filepaths):
            app("File: %s" % filepath)
            for key, arr in od.items():
                lines.append("\tkey: `%s` --> array shape: %s" % (key, str(arr.shape)))

        return "\n".join(lines)

    def add_file(self, filepath: str) -> None:
        """Add data from `filepath`"""
        with open(filepath, "rt") as fh:
            self.odlist.append(_generic_parser_fh(fh))
            self.filepaths.append(filepath)

    @add_fig_kwargs
    def plot(self, use_index=False, fontsize=8, colormap="viridis", **kwargs):
        """
        Plot all arrays. Use multiple axes if datasets.

        Args:
            use_index: By default, the x-values are taken from the first column.
                If use_index is False, the x-values are the row index.
            fontsize: fontsize for title.
            colormap: matplotlib color map.
            kwargs: options passed to ``ax.plot``.

        Return: |matplotlib-figure|
        """
        if not self.odlist: return None

        # Compute intersection of all keys.
        # Here we loose the initial ordering in the dict but oh well!
        klist = [list(d.keys()) for d in self.odlist]
        keys = set(klist[0]).intersection(*klist)
        if not keys:
            print("Warning: cannot find common keys in files. Check input data")
            return None

        # Build grid of plots.
        num_plots, ncols, nrows = len(keys), 1, 1
        if num_plots > 1:
            ncols = 2
            nrows = (num_plots // ncols) + (num_plots % ncols)

        ax_list, fig, plt = get_axarray_fig_plt(None, nrows=nrows, ncols=ncols,
                                                sharex=False, sharey=False, squeeze=False)
        ax_list = ax_list.ravel()

        # Don't show the last ax if num_plots is odd.
        if num_plots % ncols != 0: ax_list[-1].axis("off")

        cmap = plt.get_cmap(colormap)
        line_cycle = itertools.cycle(["-", ":", "--", "-.",])

        # One ax for key, each ax may show multiple arrays
        # so we need different line styles that are consistent with input data.
        # Figure may be crowded but it's difficult to do better without metadata
        # so I'm not gonna spend time to implement more complicated logic.
        for ax, key in zip(ax_list, keys):
            ax.set_title(key, fontsize=fontsize)
            ax.grid(True)
            for iod, (od, filepath) in enumerate(zip(self.odlist, self.filepaths)):
                if key not in od: continue
                arr = od[key]
                color = cmap(iod / len(self.odlist))
                xvals = arr[0] if not use_index else list(range(len(arr[0])))
                arr_list = arr[1:] if not use_index else arr
                for iarr, (ys, linestyle) in enumerate(zip(arr_list, line_cycle)):
                    ax.plot(xvals, ys, color=color, linestyle=linestyle,
                            label=os.path.relpath(filepath) if iarr == 0 else None)

            ax.legend(loc="best", fontsize=fontsize, shadow=True)

        return fig


##########################
# Plotly helper functions
##########################

_LATEX_GREEK_TO_UNICODE = dict(
    alpha="α",
    beta="β",
    gamma="ɣ",
    delta="δ",
    epsilon="ε",
    zeta="ζ",
    eta="η",
    theta="θ",
    iota="ι",
    kappa="κ",
    #lambda="λ",
    mu="μ",
    nu="ν",
    xi="ξ",
    omicron="ο",
    pi="π",
    rho="ρ",
    sigma="σ",
    tau="τ",
    upsilon="υ",
    phi="φ",
    chi="χ",
    psi="ψ",
    omega="ω",
    # Capital case:
    Alpha="Α",
    Beta="Β",
    Gamma="Γ",
    Delta="Δ",
    Epsilon="Ε",
    Zeta="Ζ",
    Eta="Η",
    Theta="Θ",
    Iota="Ι",
    Kappa="Κ",
    Lambda="Λ",
    Mu="Μ",
    Nu="Ν",
    Xi="Ξ",
    Omicron="Ο",
    Po="Π",
    Rho="Ρ",
    Sigma="Σ",
    Tau="Τ",
    Upsilon="Υ",
    Phi="Φ",
    Chi="Χ",
    Psi="Ψ",
    Omega="Ω",
)

_LATEX_GREEK_TO_UNICODE["lambda"] = "λ"


def latex_greek_2unicode(latex: str) -> str:
    """
    Convert a single greek letter in latex notation into unicode
    """
    s = latex.replace("$", "").replace("\\", "").strip()
    return _LATEX_GREEK_TO_UNICODE[s]


def is_plotly_figure(obj: Any) -> bool:
    """Return True if obj is a plotly Figure."""
    import plotly.graph_objs as go
    return isinstance(obj, go.Figure)
    #return isinstance(obj, (go.Figure, go.FigureWidget))


class PlotlyRowColDesc:
    """
    This object specifies the position of a plotly subplot inside a grid.

    rcd: PlotlyRowColDesc object used when fig is not None to specify the (row, col) of the subplot in the grid.
    """

    @classmethod
    def from_object(cls, obj: Any) -> PlotlyRowColDesc:
        """
        Build an instance for a generic object.
        If oject is None, a simple descriptor corresponding to a (1,1) grid is returned.
        """
        if obj is None: return cls(0, 0, 1, 1)
        if isinstance(obj, cls): return obj

        # Assume list with 4 integers
        try:
            return cls(*obj)
        except Exception as exc:
            raise TypeError(f"Dont know how to convert `{type(obj)}` into `{cls}`")

    def __init__(self, py_row: int, py_col: int, nrows: int, ncols: int):
        """
        Args:
            py_row, py_col: python index of the subplot in the grid (starts from 0)
            nrows, ncols: Number of rows/cols in the grid.
        """
        self.py_row, self.py_col = (py_row, py_col)
        self.nrows, self.ncols = (nrows, ncols)
        self.iax = 1 + self.py_col + self.py_row * self.ncols
        # Note that plotly col and row start from 1.
        if nrows == 1 and ncols == 1:
            self.ply_row, self.ply_col = (None, None)
        else:
            self.ply_row, self.ply_col = (self.py_row + 1, self.py_col + 1)

    def __str__(self) -> str:
        lines = []
        app = lines.append
        app("py_rowcol: (%d, %d) in grid: (%d, %d)" % (self.py_row, self.py_col, self.nrows, self.ncols))
        app("plotly_rowcol: (%s, %s)" % (self.ply_row, self.ply_col))

        return "\n".join(lines)

    #@lazy_property
    #def rowcol_dict(self):
    #    if self.nrows == 1 and self.ncols == 1: return {}
    #    return dict(row=self.ply_row, col=self.ply_col)


def get_figs_plotly(nrows=1, ncols=1, subplot_titles=(), sharex=False, sharey=False, **fig_kw):
    """
    Helper function used in plot functions that build the `plotly` figure by calling plotly.subplots.

    Returns:
        figure: plotly graph_objects figure
        go: plotly graph_objects module.
    """
    from plotly.subplots import make_subplots
    import plotly.graph_objects as go

    fig = make_subplots(rows=int(nrows), cols=int(ncols), subplot_titles=subplot_titles, shared_xaxes=sharex,
                        shared_yaxes=sharey, **fig_kw)

    return fig, go


def get_fig_plotly(fig=None, **fig_kw):
    """
    Helper function used in plot functions that build the `plotly` figure by calling
    plotly.graph_objects.Figure if fig is None else return fig

    Returns:
        figure: plotly graph_objects figure
        go: plotly graph_objects module.
    """
    import plotly.graph_objects as go

    if fig is None:
        fig = go.Figure(**fig_kw)
        #fig = go.FigureWidget(**fig_kw)

    return fig, go


def plotly_set_lims(fig, lims, axname, iax=None):
    """
    Set the data limits for the axis ax.

    Args:
        fig: Plotly Figure.
        lims: tuple(2) for (left, right), if tuple(1) or scalar for left only, none is set.
        axname: "x" for x-axis, "y" for y-axis.
        iax: An int, use iax=n to decorate the nth axis when the fig has subplots.

    Return: (left, right)
    """
    left, right = None, None
    if lims is None: return (left, right)

    # iax = kwargs.pop("iax", 1)
    # xaxis = 'xaxis%u' % iax
    #fig.layout[xaxis].title.text = "Wave Vector"

    axis = dict(x=fig.layout.xaxis, y=fig.layout.yaxis)[axname]

    len_lims = None
    try:
        len_lims = len(lims)
    except TypeError:
        # Assume Scalar
        left = float(lims)

    if len_lims is not None:
        if len(lims) == 2:
            left, right = lims[0], lims[1]
        elif len(lims) == 1:
            left = lims[0]

    ax_range = axis.range
    if ax_range is None and (left is None or right is None):
        return None, None

    #if left is not None: ax_range[0] = left
    #if right is not None: ax_range[1] = right

    # Example: fig.update_layout(yaxis_range=[-4,4])
    k = dict(x="xaxis", y="yaxis")[axname]
    if iax:
        k= k + str(iax)
    fig.layout[k].range = [left, right]

    return left, right


_PLOTLY_DEFAULT_SHOW = [True]


def set_plotly_default_show(true_or_false: bool) -> None:
    """
    Set the default value of show in the add_plotly_fig_kwargs decorator.
    Usefule for instance when generating the sphinx gallery of plotly plots.
    """
    _PLOTLY_DEFAULT_SHOW[0] = true_or_false


def add_plotly_fig_kwargs(func):
    """
    Decorator that adds keyword arguments for functions returning plotly figures.
    The function should return either a plotly figure or None to signal some
    sort of error/unexpected event.
    See doc string below for the list of supported options.
    """
    from functools import wraps

    @wraps(func)
    def wrapper(*args, **kwargs):
        # pop the kwds used by the decorator.
        title = kwargs.pop("title", None)
        show = kwargs.pop("show", _PLOTLY_DEFAULT_SHOW[0])
        hovermode = kwargs.pop("hovermode", False)
        savefig = kwargs.pop("savefig", None)
        write_json = kwargs.pop("write_json", None)
        config = kwargs.pop("config", None)
        renderer = kwargs.pop("renderer", None)
        chart_studio = kwargs.pop("chart_studio", False)
        template = kwargs.pop("template", None)

        # Allow users to specify the renderer via shell env.
        if renderer is not None and os.getenv("PLOTLY_RENDERER", default=None) is not None:
            renderer = None

        # Call func and return immediately if None is returned.
        fig = func(*args, **kwargs)
        if fig is None:
            return fig

        # Operate on plotly figure.
        if title is not None:
            fig.update_layout(title_text=title, title_x=0.5)

        if template is not None:
            fig.update_layout(template=template)

        if savefig:
            # https://plotly.github.io/plotly.py-docs/generated/plotly.io.write_image.html
            if savefig.endswith("html"):
                from plotly.offline import plot as show_plotly
                show_plotly(fig, include_mathjax="cdn", filename=savefig, auto_open=False)

            else:
                try:
                    import kaleido
                except ImportError:
                    kaleido = False

                if kaleido is None:
                    raise ValueError(
                        "kaleido package required to save static ploty images\n"
                        "please install it using:\npip install kaleido"
                    )

                fig.write_image(savefig, engine="kaleido", scale=5, width=750, height=750)
                #fig.write_image(savefig)

        if write_json:
            import plotly.io as pio
            pio.write_json(fig, write_json)

        fig.layout.hovermode = hovermode

        if show: # and _PLOTLY_DEFAULT_SHOW:
            my_config = dict(
                responsive=True,
                #showEditInChartStudio=True,
                showLink=True,
                plotlyServerURL="https://chart-studio.plotly.com",
            )

            if config is not None:
                my_config.update(config)

            #add_template_buttons(fig)

            fig.show(renderer=renderer, config=my_config)

        if chart_studio:
            push_to_chart_studio(fig)

        return fig

    # Add docstring to the decorated method.
    s = (
            "\n\n"
            + """\
                Keyword arguments controlling the display of the figure:
                ================  ====================================================================
                kwargs            Meaning
                ================  ====================================================================
                title             Title of the plot (Default: None).
                show              True to show the figure (default: True).
                hovormode         True to show the hover info (default: False)
                savefig           "abc.png" , "abc.jpeg" or "abc.webp" to save the figure to a file.
                write_json        Write plotly figure to `write_json` JSON file.
                                  Inside jupyter-lab, one can right-click the `write_json` file from the file menu
                                  and open with "Plotly Editor".
                                  Make some changes to the figure, then use the file menu to save the customized plotly plot.
                                  Requires `jupyter labextension install jupyterlab-chart-editor`.
                                  See https://github.com/plotly/jupyterlab-chart-editor
                renderer          (str or None (default None)) –
                                  A string containing the names of one or more registered renderers
                                  (separated by ‘+’ characters) or None. If None, then the default
                                  renderers specified in plotly.io.renderers.default are used.
                                  See https://plotly.com/python-api-reference/generated/plotly.graph_objects.Figure.html
                config (dict)     A dict of parameters to configure the figure. The defaults are set in plotly.js.
                chart_studio      True to push figure to chart_studio server. Requires authenticatios.
                                  Default: False.
                template          Plotly template. See https://plotly.com/python/templates/
                                  ["plotly", "plotly_white", "plotly_dark", "ggplot2", "seaborn", "simple_white", "none"]
                                  Default is None that is the default template is used.
                ================  ====================================================================
        """
    )

    if wrapper.__doc__ is not None:
        # Add s at the end of the docstring.
        wrapper.__doc__ += "\n" + s
    else:
        # Use s
        wrapper.__doc__ = s

    return wrapper


def plotlyfigs_to_browser(figs, filename=None, browser=None):
    """
    Save a list of plotly figures in an HTML file and open it the browser.
    Useful to display multiple figures generated by different AbiPy methods
    without having to construct a plotly subplot grid.

    Args:
        figs: List of plotly figures.
        filename: File name to save in. Use temporary filename if filename is None.
        browser: Open webpage in ``browser``. Use $BROWSER if None.

    Example:

        fig1 = plotter.combiplotly(renderer="browser", title="foo", show=False)
        fig2 = plotter.combiplotly(renderer="browser", title="bar", show=False)
        from abipy.tools.plotting import plotlyfigs_to_browser
        plotlyfigs_to_browser([fig1, fig2])

    Return: path to HTML file.
    """
    if filename is None:
        import tempfile
        fd, filename = tempfile.mkstemp(text=True, suffix=".html")

    if not isinstance(figs, (list, tuple)): figs = [figs]

    # Based on https://stackoverflow.com/questions/46821554/multiple-plotly-plots-on-1-page-without-subplot
    with open(filename, "wt") as fp:
        for i, fig in enumerate(figs):
            first = True if i == 0 else False
            fig.write_html(fp, include_plotlyjs=first, include_mathjax="cdn" if first else False)

    import webbrowser
    print("Opening HTML file:", filename)
    webbrowser.get(browser).open_new_tab("file://" + filename)

    return filename


def plotly_klabels(labels, allow_dupes=False):
    """
    This helper function polish a list of k-points labels before calling plotly by:

        - Checking if we have two equivalent consecutive labels (only the first one is shown and the second one is set to "")
        - Replacing particular Latex tokens with unicode as plotly support for Latex is far from optimal.

    Return: New list labels, same length as input labels.
    """
    new_labels = labels.copy()

    if not allow_dupes:
        # Don't show label if previous k-point is the same.
        for il in range(1, len(new_labels)):
            if new_labels[il] == new_labels[il - 1]: new_labels[il] = ""

    replace = {
        r"$\Gamma$": "Γ",
    }

    for il in range(len(new_labels)):
        if new_labels[il] in replace:
            new_labels[il] = replace[new_labels[il]]

    return new_labels


def plotly_set_xylabels(fig, xlabel, ylabel, exchange_xy):
    """
    Set the x- and the y-label of axis ax, exchanging x and y if exchange_xy
    """
    if exchange_xy: xlabel, ylabel = ylabel, xlabel
    fig.layout.xaxis.title.text = xlabel
    fig.layout.yaxis.title.text = ylabel


_PLOTLY_AUTHEHTICATED = False


def plotly_chartstudio_authenticate():
    """
    Authenticate the user on the chart studio portal by reading `PLOTLY_USERNAME` and `PLOTLY_API_KEY`
    from the pymatgen configuration file located in $HOME/.pmgrc.yaml.

        PLOTLY_USERNAME: johndoe
        PLOTLY_API_KEY: XXXXXXXXXXXXXXXXXXXX

    """
    global _PLOTLY_AUTHEHTICATED
    if _PLOTLY_AUTHEHTICATED: return

    try:
        from pymatgen.core import SETTINGS
        #from pymatgen.settings import SETTINGS
    except ImportError:
        from pymatgen import SETTINGS

    example = """
Add it to $HOME/.pmgrc.yaml using the follow syntax:

PLOTLY_USERNAME: john_doe
PLOTLY_API_KEY: secret  # to get your api_key go to profile > settings > regenerate key

"""

    username = SETTINGS.get("PLOTLY_USERNAME")
    if username is None:
        raise RuntimeError(f"Cannot find PLOTLY_USERNAME in pymatgen settings.\n{example}")

    api_key = SETTINGS.get("PLOTLY_API_KEY")
    if api_key is None:
        raise RuntimeError(f"Cannot find PLOTLY_API_KEY in pymatgen settings.\n{example}")

    import chart_studio
    # https://towardsdatascience.com/how-to-create-a-plotly-visualization-and-embed-it-on-websites-517c1a78568b
    chart_studio.tools.set_credentials_file(username=username, api_key=api_key)
    _PLOTLY_AUTHEHTICATED = True


def push_to_chart_studio(figs):
    """
    Push a plotly figure or a list of figures to the chart studio cloud.
    """
    plotly_chartstudio_authenticate()
    import chart_studio.plotly as py
    if not isinstance(figs, (list, tuple)): figs = [figs]
    for fig in figs:
        py.plot(fig, auto_open=True)


####################################################
# This code is shamelessy taken from Adam's package
####################################################
#import plotly.graph_objects as go


def go_points(points, size=4, color="black", labels=None, **kwargs):

    #textposition = 'top right',
    #textfont = dict(color='#E58606'),
    mode = "markers" if labels is None else "markers+text"
    #text = labels

    if labels is not None:
        labels = plotly_klabels(labels, allow_dupes=True)

    import plotly.graph_objects as go
    return go.Scatter3d(
        x=[v[0] for v in points],
        y=[v[1] for v in points],
        z=[v[2] for v in points],
        marker=dict(size=size, color=color),
        mode=mode,
        text=labels,
        **kwargs
    )


def _add_if_not_in(d, key, value):
    if key not in d:
        d[key] = value


def go_line(v1, v2, color="black", width=2, mode="lines", **kwargs):

    _add_if_not_in(kwargs, "line_color", "black")
    _add_if_not_in(kwargs, "line_width", 2)

    import plotly.graph_objects as go
    return go.Scatter3d(
        mode=mode,
        x=[v1[0], v2[0]],
        y=[v1[1], v2[1]],
        z=[v1[2], v2[2]],
        #line=dict(color=color, width=width),
        **kwargs
    )


def go_lines(V, name=None, color="black", width=2, **kwargs):
    import plotly.graph_objects as go
    gen = ((v1, v2) for (v1, v2) in V)
    v1, v2 = next(gen)
    out = [
        go_line(v1, v2, width=width, color=color, name=name, legendgroup=name, **kwargs)
    ]
    out.extend(
        go_line(
            v1,
            v2,
            width=width,
            color=color,
            showlegend=False,
            legendgroup=name,
            **kwargs
        )
        for (v1, v2) in gen
    )
    return out


def vectors(lattice, name=None, color="black", width=4, **kwargs):
    gen = zip(lattice, ["a", "b", "c"])
    v, label = next(gen)

    out = [
        go_line(
            [0, 0, 0],
            v,
            text=["", label],
            width=width,
            color=color,
            name=name,
            legendgroup=name,
            mode="lines+text",
            **kwargs
        )
    ]
    out.extend(
        go_line(
            [0, 0, 0],
            v,
            text=["", label],
            width=width,
            color=color,
            showlegend=False,
            legendgroup=name,
            mode="lines+text",
            **kwargs
        )
        for (v, label) in gen
    )
    return out


def get_vectors(lattice_mat, name=None, color="black", width=2, **kwargs):
    return go_lines([[[0, 0, 0], v] for v in lattice_mat], **kwargs)


def get_box(lattice_mat, **kwargs):
    a, b, c = lattice_mat
    segments = [
        [[0, 0, 0], a],
        [[0, 0, 0], b],
        [[0, 0, 0], c],
        [a, a + b],
        [a, a + c],
        [b, b + a],
        [b, b + c],
        [c, c + a],
        [c, c + b],
        [a + b, a + b + c],
        [a + c, a + b + c],
        [b + c, a + b + c],
    ]
    return go_lines(segments, **kwargs)


def plot_fcc_conv():

    fcc_conv = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
    fcc_vectors = vectors(
        fcc_conv, name="conv lattice vectors", color="darkblue", width=6
    )
    fcc_box = get_box(fcc_conv, name="conv lattice")

    atoms = go_points(
        [[0, 0, 0], [0.5, 0.5, 0], [0.5, 0, 0.5], [0, 0.5, 0.5]],
        size=10,
        color="orange",
        name="atoms",
        legendgroup="atoms",
    )

    import plotly.graph_objects as go
    fig = go.Figure(data=[*fcc_box, *fcc_vectors, atoms])
    return fig


def plot_fcc_prim():
    fcc_prim = np.array([[0.5, 0.5, 0], [0, 0.5, 0.5], [0.5, 0, 0.5]])

    fcc_prim_vectors = vectors(
        fcc_prim, name="prim lattice vectors", color="green", width=6
    )
    fcc_prim_box = get_box(fcc_prim, name="prim lattice", color="green")

    atoms = go_points(
        [[0, 0, 0], [0.5, 0.5, 0], [0.5, 0, 0.5], [0, 0.5, 0.5]],
        size=10,
        color="orange",
        name="atoms",
        legendgroup="atoms",
    )

    fcc_conv = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
    fcc_conv_box = get_box(fcc_conv, name="conv lattice")

    import plotly.graph_objects as go
    fig = go.Figure(data=[*fcc_prim_box, *fcc_prim_vectors, *fcc_conv_box, atoms])

    return fig


def plot_fcc_100():

    # fcc_100_cell = np.array([[0, 0.5, -0.5], [0, 0.5, 0.5], [1.0, 0.0, 0]])
    fcc_100_cell = np.array([[0.5, -0.5, 0], [0.5, 0.5, 0], [0.0, 0, 1.0]])

    fcc_100_vectors = vectors(
        fcc_100_cell, name="100 lattice vectors", color="red", width=6
    )
    fcc_100_box = get_box(fcc_100_cell, name="100 lattice", color="red")

    fig = plot_fcc_conv()
    fig.add_traces([*fcc_100_box, *fcc_100_vectors])

    return fig


def plot_fcc_110():
    fcc_110_cell = np.array([[0, 0.0, 1.0], [0.5, -0.5, 0], [0.5, 0.5, 0.0]])

    fcc_110_vectors = vectors(
        fcc_110_cell, name="reduced lattice vectors", color="red", width=6
    )
    fcc_110_box = get_box(fcc_110_cell, name="reduced lattice", color="red")

    fig = plot_fcc_conv()
    fig.add_traces([*fcc_110_box, *fcc_110_vectors])
    return fig


def plot_fcc_111():
    fcc_111_cell = np.array([[0.5, 0, -0.5], [0, 0.5, -0.5], [1, 1, 1]])

    fcc_111_vectors = vectors(
        fcc_111_cell, name="reduced lattice vectors", color="red", width=6
    )
    fcc_111_box = get_box(fcc_111_cell, name="reduced lattice", color="red")

    fig = plot_fcc_conv()
    fig.add_traces([*fcc_111_box, *fcc_111_vectors])
    return fig


def plotly_structure(structure, ax=None, to_unit_cell=False, alpha=0.7,
                     style="points+labels", color_scheme="VESTA", **kwargs):
    """
    Plot structure with plotly (minimalistic version).

    Args:
        structure: |Structure| object
        ax: matplotlib :class:`Axes3D` or None if a new figure should be created.
        alpha: The alpha blending value, between 0 (transparent) and 1 (opaque)
        to_unit_cell: True if sites should be wrapped into the first unit cell.
        style: "points+labels" to show atoms sites with labels.
        color_scheme: color scheme for atom types. Allowed values in ("Jmol", "VESTA")

    Returns: |matplotlib-Figure|
    """
    #fig, ax = plot_unit_cell(structure.lattice, ax=ax, linewidth=1)

    box = get_box(structure.lattice.matrix) #, **kwargs):

    from pymatgen.analysis.molecule_structure_comparator import CovalentRadius
    from pymatgen.vis.structure_vtk import EL_COLORS

    #symb2data = {}
    #for symbol in structure.symbol_set:
    #    symb2data[symbol] = d = {}
    #    d["color"] = color = tuple(i / 255 for i in EL_COLORS[color_scheme][symbol])
    #    d["radius"] = CovalentRadius.radius[symbol]
    #    inds = structure.indices_from_symbol(symbol)
    #    sites = [structure[i] for i in inds]
    #    d["xyz"] = []
    #    for site in sites:
    #       if to_unit_cell and hasattr(site, "to_unit_cell"): site = site.to_unit_cell()
    #       Use cartesian coordinates.
    #       x, y, z = site.coords
    #       d["xyz"].append((x, y ,z)

    xyz, sizes, colors = np.empty((len(structure), 3)), [], []
    for i, site in enumerate(structure):
        symbol = site.specie.symbol
        color = tuple(i / 255 for i in EL_COLORS[color_scheme][symbol])
        radius = CovalentRadius.radius[symbol]
        if to_unit_cell and hasattr(site, "to_unit_cell"): site = site.to_unit_cell()
        # Use cartesian coordinates.
        x, y, z = site.coords
        xyz[i] = (x, y, z) # , radius)
        sizes.append(radius)
        colors.append(color)
        #if "labels" in style:
        #    ax.text(x, y, z, symbol)

    atoms = go_points(
        #[[0, 0, 0], [0.5, 0.5, 0], [0.5, 0, 0.5], [0, 0.5, 0.5]],
        xyz,
        size=10,
        color="orange",
        name="atoms",
        legendgroup="atoms",
    )

    #marker = [dict(size=size, color=color) for (size, color) in zip(sizes, colors)]

    #atoms = go.Scatter3d(
    #    x=[v[0] for v in xyz],
    #    y=[v[1] for v in xyz],
    #    z=[v[2] for v in xyz],
    #    #marker=dict(size=size, color=color),
    #    marker=marker,
    #    mode="markers",
    #    #**kwargs
    #)

    # The definition of sizes is not optimal because matplotlib uses points
    # whereas we would like something that depends on the radius (5000 seems to give reasonable plots)
    # For possibile approaches, see
    # https://stackoverflow.com/questions/9081553/python-scatter-plot-size-and-style-of-the-marker/24567352#24567352
    # https://gist.github.com/syrte/592a062c562cd2a98a83
    #if "points" in style:
    #    x, y, z, s = xyzs.T.copy()
    #    s = 5000 * s ** 2
    #    ax.scatter(x, y, zs=z, s=s, c=colors, alpha=alpha)  #facecolors="white", #edgecolors="blue"

    #ax.set_title(structure.composition.formula)
    #ax.set_axis_off()

    #fig = go.Figure(data=[*box, *vectors, atoms])
    import plotly.graph_objects as go
    fig = go.Figure(data=[*box, atoms])
    return fig


# This is the matplotlib API to plot the BZ.

def plotly_wigner_seitz(lattice, fig=None, **kwargs):
    """
    Adds the skeleton of the Wigner-Seitz cell of the lattice to a plotly figure.

    Args:
        lattice: Lattice object
        fig: plotly figure or None if a new figure should be created.
        kwargs: kwargs passed to the matplotlib function 'plot'. Color defaults to black
            and linewidth to 1.

    Returns: Plotly figure
    """
    #ax, fig, plt = get_ax3d_fig_plt(ax)
    fig, go = get_fig_plotly(fig=fig) #, **fig_kw)

    if "line_color" not in kwargs:
        kwargs["line_color"] = "black"
    if "line_width" not in kwargs:
        kwargs["line_width"] = 1

    bz = lattice.get_wigner_seitz_cell()
    #ax, fig, plt = get_ax3d_fig_plt(ax)

    for iface in range(len(bz)):  # pylint: disable=C0200
        for line in itertools.combinations(bz[iface], 2):
            for jface in range(len(bz)):
                if (
                    iface < jface
                    and any(np.all(line[0] == x) for x in bz[jface])
                    and any(np.all(line[1] == x) for x in bz[jface])
                ):
                    #ax.plot(*zip(line[0], line[1]), **kwargs)
                    fig.add_trace(go_line(line[0], line[1], showlegend=False, **kwargs))

    return fig


def plotly_lattice_vectors(lattice, fig=None, **kwargs):
    """
    Adds the basis vectors of the lattice provided to a plotly figure.

    Args:
        lattice: Lattice object
        fig: plotly figure or None if a new figure should be created.
        kwargs: kwargs passed to the matplotlib function 'plot'. Color defaults to green
            and linewidth to 3.

    Returns: plotly figure
    """
    fig, go = get_fig_plotly(fig=fig)

    if "line_color" not in kwargs:
        kwargs["line_color"] = "green"
    if "line_width" not in kwargs:
        kwargs["line_width"] = 3
    if "showlegend" not in kwargs:
        kwargs["showlegend"] = False

    vertex1 = lattice.get_cartesian_coords([0.0, 0.0, 0.0])
    vertex2 = lattice.get_cartesian_coords([1.0, 0.0, 0.0])
    fig.add_trace(go_line(vertex1, vertex2, name="a", **kwargs))
    vertex2 = lattice.get_cartesian_coords([0.0, 1.0, 0.0])
    fig.add_trace(go_line(vertex1, vertex2, name="b", **kwargs))
    vertex2 = lattice.get_cartesian_coords([0.0, 0.0, 1.0])
    fig.add_trace(go_line(vertex1, vertex2, name="c", **kwargs))

    return fig


def plotly_path(line, lattice=None, coords_are_cartesian=False, fig=None, **kwargs):
    """
    Adds a line passing through the coordinates listed in 'line' to a plotly figure.

    Args:
        line: list of coordinates.
        lattice: Lattice object used to convert from reciprocal to cartesian coordinates
        coords_are_cartesian: Set to True if you are providing
            coordinates in cartesian coordinates. Defaults to False.
            Requires lattice if False.
        fig: plotly figure or None if a new figure should be created.
        kwargs: kwargs passed to the matplotlib function 'plot'. Color defaults to red
            and linewidth to 3.

    Returns: plotly figure
    """
    fig, go = get_fig_plotly(fig=fig)

    if "line_color" not in kwargs:
        kwargs["line_color"] = "red"
    if "line_width" not in kwargs:
        kwargs["line_width"] = 3

    for k in range(1, len(line)):
        vertex1 = line[k - 1]
        vertex2 = line[k]
        if not coords_are_cartesian:
            if lattice is None:
                raise ValueError("coords_are_cartesian False requires the lattice")
            vertex1 = lattice.get_cartesian_coords(vertex1)
            vertex2 = lattice.get_cartesian_coords(vertex2)

        fig.add_trace(go_line(vertex1, vertex2, showlegend=False, **kwargs))

    return fig


#def plotly_labels(labels, lattice=None, coords_are_cartesian=False, ax=None, **kwargs):
#    """
#    Adds labels to a matplotlib Axes
#
#    Args:
#        labels: dict containing the label as a key and the coordinates as value.
#        lattice: Lattice object used to convert from reciprocal to cartesian coordinates
#        coords_are_cartesian: Set to True if you are providing.
#            coordinates in cartesian coordinates. Defaults to False.
#            Requires lattice if False.
#        ax: matplotlib :class:`Axes` or None if a new figure should be created.
#        kwargs: kwargs passed to the matplotlib function 'text'. Color defaults to blue
#            and size to 25.
#
#    Returns:
#        matplotlib figure and matplotlib ax
#    """
#    ax, fig, plt = get_ax3d_fig_plt(ax)
#
#    if "color" not in kwargs:
#        kwargs["color"] = "b"
#    if "size" not in kwargs:
#        kwargs["size"] = 25
#
#    for k, coords in labels.items():
#        label = k
#        if k.startswith("\\") or k.find("_") != -1:
#            label = "$" + k + "$"
#        off = 0.01
#        if coords_are_cartesian:
#            coords = np.array(coords)
#        else:
#            if lattice is None:
#                raise ValueError("coords_are_cartesian False requires the lattice")
#            coords = lattice.get_cartesian_coords(coords)
#        ax.text(*(coords + off), s=label, **kwargs)
#
#    return fig, ax


def plotly_points(points, lattice=None, coords_are_cartesian=False, fold=False, labels=None, fig=None, **kwargs):
    """
    Adds points to a plotly figure.

    Args:
        points: list of coordinates
        lattice: Lattice object used to convert from reciprocal to cartesian coordinates
        coords_are_cartesian: Set to True if you are providing
            coordinates in cartesian coordinates. Defaults to False.
            Requires lattice if False.
        fold: whether the points should be folded inside the first Brillouin Zone.
            Defaults to False. Requires lattice if True.
        fig: plotly figure or None if a new figure should be created.
        kwargs: kwargs passed to the matplotlib function 'scatter'. Color defaults to blue

    Returns: plotly figure
    """
    fig, go = get_fig_plotly(fig=fig) #, **fig_kw)

    if "marker_color" not in kwargs:
        kwargs["marker_color"] = "blue"

    if (not coords_are_cartesian or fold) and lattice is None:
        raise ValueError("coords_are_cartesian False or fold True require the lattice")

    from pymatgen.electronic_structure.plotter import fold_point
    vecs = []
    for p in points:

        if fold:
            p = fold_point(p, lattice, coords_are_cartesian=coords_are_cartesian)

        elif not coords_are_cartesian:
            p = lattice.get_cartesian_coords(p)

        vecs.append(p)

    kws = dict(textposition="top right", showlegend=False) #, textfont=dict(color='#E58606'))
    kws.update(kwargs)
    fig.add_trace(go_points(vecs, labels=labels, **kws))

    return fig


@add_plotly_fig_kwargs
def plotly_brillouin_zone_from_kpath(kpath, fig=None, **kwargs):
    """
    Gives the plot (as a matplotlib object) of the symmetry line path in
        the Brillouin Zone.

    Args:
        kpath (HighSymmKpath): a HighSymmKPath object
        ax: matplotlib :class:`Axes` or None if a new figure should be created.
        **kwargs: provided by add_fig_kwargs decorator

    Returns: plotly figure.
    """
    lines = [[kpath.kpath["kpoints"][k] for k in p] for p in kpath.kpath["path"]]
    return plotly_brillouin_zone(
        bz_lattice=kpath.prim_rec,
        lines=lines,
        fig=fig,
        labels=kpath.kpath["kpoints"],
        show=False,
        **kwargs,
    )


@add_plotly_fig_kwargs
def plotly_brillouin_zone(
    bz_lattice,
    lines=None,
    labels=None,
    kpoints=None,
    fold=False,
    coords_are_cartesian=False,
    fig=None,
    **kwargs,
):
    """
    Plots a 3D representation of the Brillouin zone of the structure.
    Can add to the plot paths, labels and kpoints

    Args:
        bz_lattice: Lattice object of the Brillouin zone
        lines: list of lists of coordinates. Each list represent a different path
        labels: dict containing the label as a key and the coordinates as value.
        kpoints: list of coordinates
        fold: whether the points should be folded inside the first Brillouin Zone.
            Defaults to False. Requires lattice if True.
        coords_are_cartesian: Set to True if you are providing
            coordinates in cartesian coordinates. Defaults to False.
        ax: matplotlib :class:`Axes` or None if a new figure should be created.
        kwargs: provided by add_fig_kwargs decorator

    Returns: plotly figure
    """

    fig = plotly_lattice_vectors(bz_lattice, fig=fig)
    plotly_wigner_seitz(bz_lattice, fig=fig)
    if lines is not None:
        for line in lines:
            plotly_path(line, bz_lattice, coords_are_cartesian=coords_are_cartesian, fig=fig)

    if labels is not None:
        # TODO
        #plotly_labels(labels, bz_lattice, coords_are_cartesian=coords_are_cartesian, ax=ax)
        plotly_points(
            labels.values(),
            lattice=bz_lattice,
            coords_are_cartesian=coords_are_cartesian,
            fold=False,
            labels=list(labels.keys()),
            fig=fig,
        )

    if kpoints is not None:
        plotly_points(
            kpoints,
            lattice=bz_lattice,
            coords_are_cartesian=coords_are_cartesian,
            fold=fold,
            fig=fig,
        )

    #ax.set_xlim3d(-1, 1)
    #ax.set_ylim3d(-1, 1)
    #ax.set_zlim3d(-1, 1)
    # ax.set_aspect('equal')
    #ax.axis("off")

    return fig


def add_colorscale_dropwdowns(fig):
    """
    Add dropdown widgets to change/reverse the colorscale.
    Based on: https://plotly.com/python/dropdowns/#update-several-data-attributes
    """
    button_layer_1_height = 1.30

    # Create list of buttons
    # A single button has the form:
    #
    #    dict(
    #        args=["colorscale", "Viridis"],
    #        label="Viridis",
    #        method="restyle"
    #    ),

    colorscales = ["Viridis", "Cividis", "Blues", "Greens"]

    colorscale_buttons = []
    for cscale in colorscales:
        colorscale_buttons.append(dict(
                args=["colorscale", cscale],
                label=cscale,
                method="restyle",
        ))

    fig.update_layout(
        updatemenus=[
            dict(
                buttons=colorscale_buttons,
                direction="down",
                pad={"r": 10, "t": 10},
                showactive=True,
                x=0.1,
                xanchor="left",
                y=button_layer_1_height,
                yanchor="top"
            ),
            dict(
                buttons=list([
                    dict(
                        args=["reversescale", False],
                        label="False",
                        method="restyle"
                    ),
                    dict(
                        args=["reversescale", True],
                        label="True",
                        method="restyle"
                    )
                ]),
                direction="down",
                pad={"r": 10, "t": 10},
                showactive=True,
                x=0.37,
                xanchor="left",
                y=button_layer_1_height,
                yanchor="top"
            ),
        ]
    )

    y = button_layer_1_height - 0.02

    fig.update_layout(
        annotations=[
            dict(text="colorscale", x=0, xref="paper", y=y, yref="paper",
                 align="left", showarrow=False),
            dict(text="Reverse<br>Colorscale", x=0.25, xref="paper", y=y,
                 yref="paper", showarrow=False),
    ])

    return fig



"""
"""
from __future__ import annotations

#from typing import List  # Dict, Tuple
from abc import ABC, abstractmethod
from pydantic import BaseModel, Field, validator
from bson.objectid import ObjectId
from pymatgen.io.abinit.pseudos import Pseudo, PseudoTable
from abipy.core.structure import Structure
from abipy.flowtk.psrepos import get_repo_from_name
from .base_models import AbipyModel, GfsFileDesc, GfsDesc, MongoConnector


#class PseudoSpecs(_PseudosProvider):
class _PseudosProvider(BaseModel, ABC):
    """
    Abstract base class enforcing the implementation of ``get_pseudos``
    """

    ps_type: str = Field(..., description="Pseudopotential type e.g. NC or PAW.")

    ps_generator: str = Field(..., description="Name of the Pseudopotential generator.")

    xc_name: str = Field(..., description="Name of the XC functional e.g. LDA, PBE, PBEsol.")

    relativity_type: str = Field(..., description="Scalar-relativistic (SR) or fully-relativistic (FR)")

    @abstractmethod
    def get_pseudos(self) -> PseudoTable:
        """Build and return the PseudoTable associated the model."""


#class PseudoSpecsWithRepo(_PseudosProvider):
class PseudoSpecs(_PseudosProvider):
    """
    This model stores the parameters needed to retrieve the PseudoPotential Table from a repository.
    The `repo_name` defines the full set of versioned pseudos while `table_name` defines the list of pseudos.
    included in the final Table.
    Note that the PseudoDojo/JTC project may provide multiple pseudos for the same atom.
    This is the reason why the final table with one pseudo per element is uniquely defined by
    the tuple (repo_name, table_name).

    .. code-block::

        specs = PseudoSpecs.from_repo_table_name("ONCVPSP-PBEsol-SR-PDv0.4", "standard")
    """

    repo_name: str = Field(..., description="Name of the pseudopotential repository.")

    table_name: str = Field(..., description="Name of the table e.g. normal, stringent, MD ...")

    project_name: str = Field(..., description="PD for PseudoDojo, JTH for JTH.")

    version: str = Field(..., description="Version of the repository.")

    @classmethod
    def from_repo_table_name(cls, repo_name: str, table_name: str) -> PseudoSpecs:
        """
        Build the object from a repository name and ``table_name``.
        """
        repo = get_repo_from_name(repo_name)
        data = dict(repo_name=repo_name, table_name=table_name)
        attr_list = ["ps_generator", "ps_type", "xc_name", "relativity_type", "project_name", "version"]
        data.update({aname: getattr(repo, aname) for aname in attr_list})

        return cls(**data)

    @validator("relativity_type")
    def validate_relativity_type(cls, value):
        """
        Validate relativity_type
        """
        allowed_values = ("SR", "FR")
        if value not in allowed_values:
            raise ValueError(f"{value} not in {allowed_values}")

        return value

    def get_pseudos(self) -> PseudoTable:
        """
        Build and return the PseudoTable associated to the specs.
        """
        repo = get_repo_from_name(self.repo_name)
        if not repo.is_installed():
            print(f"Warning: Could not find {self.repo_name} installed in {repo.dirpath}.\n"
                  f"Will try to install it at runtime...")
            repo.install(verbose=1)

        # Note that the repo instance keeps an internal cache of tables.
        return repo.get_pseudos(table_name=self.table_name)


class MongoDbPseudos(_PseudosProvider):
    """
    A rather advanced provider that takes pseudopotentials by IDs from a MongoDB collection.
    This approach is used for testing/validanting "unofficial" pseudos.
    """

    oid_list: List[ObjectId] = Field(..., description="List of ids in collection")

    mng_connector: MongoConnector = Field(..., description="")

    #pseudo_model_cls =

    def get_pseudos(self) -> PseudoTable:
        """
        Build and return the PseudoTable associated to the specs.
        """
        collection = self.mng_connector.get_collection()
        #pseudo_model = pseudo_model_cls(**d)
        pseudos = []
        for oid in self.oid_list:
            d = collection.find_one({"_id": oid}, {"_id": 0})
            if d is None:
                raise ValueError(f"Cannot find pseudo with id {oid} in collection: {collection}")
            #pseudo_model = pseudo_model_cls(**d)
            p = pseudo_model.get_pseudo_from_db(self.mng_connector)
            pseudos.append(p)

        return PseudoTable(pseudos)



class BasePseudoModel(AbipyModel, ABC):
    """
    Abstract base class for a model associated to a single pseudopotential.
    Used for fetching the pseudopotential file from a MongoDB collection.
    """

    atomic_symbol: str = Field(..., description="")

    ps_type: str = Field(..., description="Pseudopotential type e.g. NC or PAW.")

    ps_generator: str = Field(..., description="Name of the Pseudopotential generator.")

    xc_name: str = Field(..., description="Name of the XC functional e.g. LDA, PBE, PBEsol.")

    relativity_type: str = Field(..., description="Scalar-relativistic (SR) or fully-relativistic (FR)")

    #psgen_input_string: str = Field(..., description="")

    #psgen_output_string: str = Field(..., description="")

    #output_gsfd: GfsFileDesc = Field(..., description="Pointer to the main output file in GridFs")

    @abstractmethod
    def get_pseudo_from_db(self, mng_connector: MongoConnector) -> Pseudo:
        """Build and return a Pseudo object from the model."""



class OncvpspModel(BasePseudoModel):
    """
    This model stores the pseudopotentials generated by oncvpsp with the corresponding
    input and output files.
    """

    psp8_gfsd: GfsFileDesc = Field(..., description="Pointer to the psp8 file in GridFs")
    psp8_md5: str = Field(..., description="md5 checksum for the psp8 file")

    upf_md5: str = Field(..., description="md5 checksum for the upf file")
    upf_gsfd: GfsFileDesc = Field(..., description="Pointer to the UPF file in GridFs")

    psml_md5: str = Field(..., description="md5 checksum for the psml file")
    psml_gsfd: GfsFileDesc = Field(..., description="Pointer to the PSML file in GridFs")

    #@validator("atomic_symbol")
    #def validate_atomic_symbol(cls, value):

    @classmethod
    def from_dirpath(cls, dirpath: str, mng_connector: MongoConnector) -> OncvpspModel:
        """
        Build a the model from a directory containing the output files produced by oncvpsp.
        """
        from abipy.flowtk.psrepos import md5_for_filepath
        dirpath = os.path.abspath(dirpath)
        paths = [os.path.join(dirpath, b) for b in os.listdir(dirpath)]

        def find_abspath_byext(ext: str):
            cnt, ret_apath = 0, None
            for p in paths:
                if p.endswith(ext):
                    ret_apath = p
                    cnt += 1

            if cnt == 1:
                return ret_apath
            if cnt > 1:
                raise RuntimeError(f"Found multiple files with extension {ext} in {paths}")

            raise RuntimeError(f"Cannot find file with extension {ext} in {paths}")

        data = {}
        apath = find_abspath_byext(".in")
        input_string = open(apath, "rt").read()
        #data["psgen_input_string"] = input_string
        apath = find_abspath_byext(".out")
        #data["psgen_output_string"] = open(apath, "rt").read()
        #data[output_gfsd"] = mng_connector.gfs_put_filepath(apath)

        data["atomic_symbol"] = "foo"

        for fmt in ["psp8", "upf", "psml"]:
            apath = find_abspath_byext("." + fmt)
            data[f"{fmt}_md5"] = md5_for_filepath(apath)
            #with open(apath, "rt") as fh:
            #    data[f"{fmt}_input_string"] = fh.read()

            data[f"{fmt}_gfsd"] = mng_connector.gfs_put_filepath(apath)

        return cls(**data)

    def get_pseudo_from_db(self, mng_connector: MongoConnector) -> Pseudo:
        """
        Build and return a Pseudo instance from the model.
        """
        filepath = mng_connector.mktmp_filepath(self.psp8_gfsd)
        return Pseudo.from_file(filepath)




""""AbiPy panels for electronic properties."""

import param
import panel as pn
import panel.widgets as pnw

from .core import AbipyParameterized, ActiveBar, ply, mpl, depends_on_btn_click


class CompareEbandsWithMP(AbipyParameterized):

    with_gaps = param.Boolean(True)
    ylims_ev = param.Range(default=(-10, +10), doc="Energy window around the Fermi energy.")

    info_str = """
This app alllows users to upload two files with KS energies.
"""

    def __init__(self, **params):

        super().__init__(**params)

        help_md = pn.pane.Markdown(f"""
## Description

{self.info_str}
""")

        self.main_area = pn.Column(help_md,
                                   self.get_alert_data_transfer(),
                                   sizing_mode="stretch_width")

        self.replot_btn = pnw.Button(name="Replot", button_type='primary')

        self.file_input = pnw.FileInput(height=60, css_classes=["pnx-file-upload-area"])
        self.file_input.param.watch(self.on_file_input, "value")
        self.mp_progress = pn.indicators.Progress(name='Fetching data from the MP website', bar_color="warning",
                                                  active=False, width=200, height=10, align="center")

    def on_file_input(self, event):
        self.abinit_ebands = self.get_ebands_from_file_input(self.file_input)

        # Match Abinit structure with MP
        mp = self.abinit_ebands.structure.mp_match()
        if not mp.structures:
            raise RuntimeError("No structure found in the MP database")

        # Get structures from MP as AbiPy ElectronBands.
        from abipy.electrons.ebands import ElectronBands
        self.mp_ebands_list = []
        with ActiveBar(self.mp_progress):
            for mp_id in mp.ids:
                if mp_id == "this": continue
                eb = ElectronBands.from_mpid(mp_id)
                self.mp_ebands_list.append(eb)

        self.update_main()

    def update_main(self):
        col = self.pws_col(["## Plot options", "with_gaps", "ylims_ev", "replot_btn"])
        ca = col.append

        ca("## Abinit Electronic band structure:")
        ylims = self.ylims_ev
        ca(ply(self.abinit_ebands.plotly(e0="fermie", ylims=ylims, with_gaps=self.with_gaps, show=False)))

        for mp_ebands in self.mp_ebands_list:
            ca("## MP Electronic band structure:")
            ca(ply(mp_ebands.plotly(e0="fermie", ylims=ylims, with_gaps=self.with_gaps, show=False)))

        #self.main_area.objects = [col]
        self.main_area.objects = col.objects

    @depends_on_btn_click('replot_btn')
    def on_replot_btn(self):
        self.update_main()

    def get_panel(self):
        col = pn.Column(
            "## Upload a *nc* file with energies along a **k**-path (possibly a *GSR.nc* file):",
            self.get_fileinput_section(self.file_input),
            pn.Row("## Fetching data from the MP website: ", self.mp_progress,
                   sizing_mode="stretch_width"),
            sizing_mode="stretch_width")

        main = pn.Column(col, self.main_area, sizing_mode="stretch_width")

        cls, kwds = self.get_abinit_template_cls_kwds()

        return cls(main=main, title="Compare with MP Ebands", **kwds)


class SkwPanelWithFileInput(AbipyParameterized):

    info_str = """
This app allows users to upload two files with KS energies.
The first file gives the energies in the IBZ used to perform the SKW interpolation
The second file contains the enegies along a k-path.
The interpolated energies are then compared with the ab-initio ones on the k-path.
The user can change the SKW intepolation parameters to gauge the quality of the SKW fit.
"""

    def __init__(self, **params):

        super().__init__(**params)

        help_md = pn.pane.Markdown(f"""
## Description

{self.info_str}
""")

        self.main_area = pn.Column(help_md,
                                   self.get_alert_data_transfer(),
                                   sizing_mode="stretch_width")

        self.ibz_file_input = pnw.FileInput(height=60, css_classes=["pnx-file-upload-area"])
        self.ibz_file_input.param.watch(self.on_ibz_file_input, "value")
        self.ebands_ibz = None

        self.kpath_file_input = pnw.FileInput(height=60, css_classes=["pnx-file-upload-area"])
        self.kpath_file_input.param.watch(self.on_kpath_file_input, "value")
        self.ebands_kpath = None

    def on_ibz_file_input(self, event):
        self.ebands_ibz = self.get_ebands_from_file_input(self.ibz_file_input)
        self.update_main_area()

    def on_kpath_file_input(self, event):
        self.ebands_kpath = self.get_ebands_from_file_input(self.kpath_file_input)
        self.update_main_area()

    def update_main_area(self):

        if self.ebands_kpath is None or self.ebands_ibz is None: return

        # SKW interpolation
        r = self.ebands_ibz.interpolate(lpratio=5, filter_params=None)

        # Build plotter.
        plotter = self.ebands_kpath.get_plotter_with("Ab-initio", "SKW interp", r.ebands_kpath)
        mpl_pane = mpl(plotter.combiplot(**self.mpl_kwargs))

        col = pn.Column(mpl_pane, sizing_mode="stretch_width")

        self.main_area.objects = [col]

    def get_panel(self):
        col = pn.Column(
            "## Upload (or drag & drop) any *nc* file with energies in the IBZ (_possibly a *GSR.nc* file_):",
            self.get_fileinput_section(self.ibz_file_input),
            "## Upload (or drag & drop) any *nc* file with energies along a **k**-path (_possibly a *GSR.nc* file_):",
            self.get_fileinput_section(self.kpath_file_input),
            sizing_mode="stretch_width")

        main = pn.Column(col, self.main_area, sizing_mode="stretch_width")
        cls, kwds = self.get_abinit_template_cls_kwds()

        return cls(main=main, title="SKW Analyzer", **kwds)



"""Tooltips for oncv GUI."""

def oncv_tip(key):
    """Tip for variable key."""
    return _ONCVTIPS[key]


_ONCVTIPS = dict(
atsym="Atomic symbol",
z="Atomic number",
nc="Number of core states",
nv="Number of valence states",
iexc="""\
Exchange-correlation functional:
   1-Wigner
   2-Hedin-Lundquist
   3-Perdew-Wang-Ceperly-Alder
   4-Perdew-Burke-Enzerhof""",
psfile="Format of pseudopotential file, psp8 for ABINIT, upf for PWSCF",
n="Principal quantum number",
l="Angular momentum",
f="Occupancy (MUST be >0)",
lmax="Maximum angular momentum for which psp is calculated (<=3)",
rc="Core radius for this l",
ep="""\
Energy at which psp is generated (eigenvalue inserted for occupied
state in reference configuration, positive energy must be specified
for barrier-confined "scattering" state for unoccupied l <=lmax.
A small positive energy is usually  good (0.1-0.25 Ha).""",
ncon="""\
Number of constraints for pseudo wave function to match all-electron wave function at rc,
value + ncon-1 derivatives, must be between 3 and 5 ("M" in the paper, Eq.(6))""",
nbas="Number of basis functions.  Must be between ncon+2 and ncon+5 (`N` in the paper, Eqs.(4-5))",
qcut="Wave vector defining `residual energy` in the RRKJ method (`q_c` in the paper, Eq.(1)",
lloc="""\
Angular momentum whose semi-local psp is taken as local.
lloc=4 denotes a smooth polynomial continuation of the all-electron potential.
If lloc<=lmax, remaining data are ignored, but must be there (zeros are OK).""",
lpopt="""\
Type of polynomial continuation for lloc=4. values 1-5  permitted.
    1) match 2 derivatives, r^0,2,4
    2) match 2 derivatives, r^0,4,6
    3) match 3 derivatives, r^0,4,5,6
    4) match 3 derivatives, r^0,4,6,8
    5) match 3 derivatives, r^0,2,4,6""",
rc5="Info not available",
dvloc0="""\
Shift of r=0 potential from basic continuation (Ha) for lloc=4
depends on lpopt above as follows, with x=(r/rc)
    1) dvloc0*(1-x^2)^3
    2) dvloc0*(1-x^4)^3
    3-5) dvloc0*(1-x^4)^4""",
nproj="Number of projectors, 1 or 2. Automatically set to 0 for l=lloc",
debl="""\
Energy added to basic psp  energy ep for 2nd projector, automatically reset to match 2nd projector with 2nd bound state
at this l when it is occupied (i.e., the psp is generated for a corresponding-l shallow core state)""",
icmod="""\
  0: no non-linear core correction charge.
  1: smooth monotonic polynomial model core charge fit at "matching" rc following reference 35.""",
fcfact="""\
Radius for above determined by  rho_core(r)=fcfact*rho_pseudo_valence(r).
Values 0.25-0.5 are usually good (look at plots)""",
epsh1="""\
Lower energy limit for `phase-shift-like` log-derivative plot,
should be below the nearest core level for this l to make sure
there are no ghosts, but -2.0 Ha usually is OK""",
epsh2="Upper energy limit, 2.0 usually good",
depsh="Energy mesh interval for plot, 0.02 usually good enough",
rcfact="rcfact scales the crossover radius to determine the range of the Teter function.",
rlmax="""\
Maximum radius for Abinit psp code 8 format output.
Must be greater than maximum rc (including lloc=4 rc), but also determines
range of diagnostic plots, so  ~2*rcmax is usually good""",
drl="""\
Mesh spacing of linear radial mesh for Abinit output.
0.02 is good for "softer" psps, 0.01 is probably better with 1st row, 3d's, or semi-core psps.""",
ncnf="""\
Number of test configurations (<=4).
The reference config is always run first as a consistency check.
Core always is the reference core configuration""",
)
#   nvcnf (repeated ncnf times) number of valence states in this test configuration
#   n, l, f  (nvcnf lines, repeated follwing nvcnf's ncnf times)
